/*
 *	Implementation of functions related to address space operations. These are
 * in particular functions for reading and writing data. We need them to make
 *	generic_file_{read,write,mmap} happy
 */



#include "sifs.h"



struct address_space_operations sifs_aops = {
	readpage:		sifs_readpage,
	prepare_write:	sifs_prepare_write,
	commit_write:	sifs_commit_write,
};



/*
 *	Retrieve a data block from the server
 */

loff_t sifs_read( struct sifs_sb_info* info, char* path, u_int8_t* pagebuf,
 loff_t count, loff_t offset )
{
	int len = 0;
	int error = 0;
	loff_t rsize = 0;
	loff_t psize = 0;
	loff_t status = 0;
	ssize_t size = 0;
	u_int32_t seqnr = 0;
	u_int8_t buffer[1056];


SIFS_PROFILE( "Read data from server:\n" );

	if( info == NULL ) {
		SIFS_DEBUG( "Private superblock info is not set\n" );
		return -EFAULT;
	}
SIFS_PROFILE( "Private superblock info is set\n" );

	memset( buffer, 0, 1056 );
	seqnr = get_seqnr( info );
SIFS_PROFILE( "Got sequence number\n" );

	if( ( len = str_encode( buffer + 30, path ) ) < 0 ) {
		SIFS_DEBUG( "Encoding path into request failed\n" );
		return len;
	}

	PUT_32BIT( buffer, seqnr );
	PUT_32BIT( buffer + 4, 18 + len );   // Packet length
	PUT_32BIT( buffer + 8, SIFS_READ );   // Opcode
	PUT_64BIT( buffer + 12, offset );   // File offset
	PUT_64BIT( buffer + 20, count );   // number of requested bytes
	PUT_16BIT( buffer + 28, 1 );   // Number of strings
SIFS_PROFILE( "Request built successfully\n" );

	down( &(info->sem) );

	if( ( size = atomic_io( ssh_write, info->socket, buffer, 30 + len ) ) < 0 ) {
		SIFS_DEBUG( "Sending request to server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Request sent to server\n" );

	if( ( size = atomic_io( ssh_read, info->socket, buffer, 12 ) ) < 0 ) {
		SIFS_DEBUG( "Receiving reply header from server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Reply header received\n" );

	if( GET_32BIT( buffer ) != seqnr ) {
		SIFS_DEBUG( "Sequence number doesn't match\n" );
		status = -EPROTO;
		goto error;
	}
SIFS_PROFILE( "Sequence number matched\n" );

	if( ( error = GET_32BIT( buffer + 8 ) ) != SIFS_EOK ) {
		SIFS_DEBUG( "Read request failed\n" );
		switch( error ) {
			case SIFS_ENAMETOOLONG:
				status = -EBADF;
				break;
			case SIFS_ENOENT:
				status = -EBADF;
				break;
			case SIFS_EIO:
				status = -EIO;
				break;
			case SIFS_EINVAL:
				status = -EINVAL;
				break;
			case SIFS_ESERVER:
			default:
				status = -EFAULT;
				break;
		}
		goto out;
	}
SIFS_PROFILE( "Read request succeeded\n" );

	if( ( status = atomic_io( ssh_read, info->socket, buffer, 8 ) ) < 0 ) {
		SIFS_DEBUG( "Receiving reply payload from server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Size of reply data received\n" );

	rsize = count;
	psize = GET_64BIT( buffer );
	if( psize < count ) {
		rsize = psize;
	}

	if( ( status = atomic_io( ssh_read, info->socket, pagebuf, rsize ) ) < 0 ) {
		SIFS_DEBUG( "Receiving reply payload from server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Reply data received\n" );

	if( psize > count ) {
		SIFS_DEBUG( "Server sent more than requested\n" );
		goto error;
	}
SIFS_PROFILE( "Requested size and sent size are equal\n" );

	goto out;

error:
	up( &(info->sem) );
	sifs_reconnect( info );
	return status;

out:
	up( &(info->sem) );
	return status;
}

/*
 *	Copy a data block from the file to the page
 */

int sifs_readpage( struct file* file, struct page* page )
{
	int error = 0;
	char path[1024];
	ssize_t result = 0;
	u_int8_t* pagebuf = kmap(page);
	loff_t count = PAGE_CACHE_SIZE;
	loff_t offset = ((loff_t) page->index) << PAGE_CACHE_SHIFT;
	struct sifs_sb_info* info = (struct sifs_sb_info*)
 file->f_dentry->d_sb->u.generic_sbp;
	struct dentry* dentry = file->f_dentry;


SIFS_PROFILE( "Read full page from server:\n" );

	if( ( error = sifs_getname( dentry, path ) ) < 0 ) {
		SIFS_DEBUG( "Rebuilding path from dentries failed\n" );
		return error;
	}
SIFS_PROFILE( "Rebuilding path from dentries succeeded\n" );

	do {
		if( ( result = sifs_read( info, path, pagebuf, count, offset ) ) < 0 ){
			SIFS_DEBUG( "Reading partitial file content failed\n" );
			goto io_error;
		}
SIFS_PROFILE( "read succeeded\n" );

		count -= result;
		offset += result;
		pagebuf += result;
		dentry->d_inode->i_atime = CURRENT_TIME;

		if( result == 0 ) {
			break;
		}
	} while( count > 0 );

	memset( pagebuf, 0, count );

	flush_dcache_page( page );
	SetPageUptodate( page );

	if( PageError( page ) ) {
		ClearPageError( page );
	}

	result = 0;

io_error:
	kunmap( page );
	UnlockPage( page );
	return result;
}



/*
 * Map data from user space into kernel space
 */

int sifs_prepare_write( struct file* file, struct page* page, unsigned int offset, unsigned int to )
{
SIFS_PROFILE( "Get data from user space\n" );
	return -EROFS;
}



/*
 *	Write user space data to server
 */

int sifs_commit_write( struct file* file, struct page* page, unsigned int offset, unsigned int to )
{
SIFS_PROFILE( "Write data to server\n" );
	return -EROFS;
}
