/*
 *	Implemetation of all file_operations related functions.
 *	This includes files and directories (which are also a kind of file)
 */



#include "sifs.h"



struct file_operations sifs_dir_ops = {
	read:		generic_read_dir,
	readdir:	sifs_readdir,
};

struct file_operations sifs_file_ops = {
	read:		generic_file_read,
	write:		generic_file_write,
	mmap:	generic_file_mmap,
	open:		sifs_open,
	release:	sifs_release,
	fsync:		sifs_fsync,
	lock:		sifs_lock,
};



/*
 *	Read the contents of a directory from the server
 */

int sifs_readdir( struct file* file, void* dirent, filldir_t filldir )
{
	int len = 0;
	int pos = 0;
	int error = 0;
	int status = 0;
	int numstr = 0;
	size_t bytes = 0;
	ssize_t size = 0;
	u_int32_t ino = 0;
	u_int32_t seqnr = 0;
	u_int32_t packetlen = 0;
	u_int8_t buffer[1040];
	char path[1024];
	char* filename;
	struct sifs_sb_info* info = NULL;
	struct dentry* dentry = file->f_dentry;
	struct inode* inode = file->f_dentry->d_inode;


SIFS_PROFILE( "Read contents of directory:\n" );

	if( !( file->f_dentry && file->f_dentry->d_sb && file->f_dentry->d_sb->u.generic_sbp ) ) {
		SIFS_DEBUG( "Private superblock info not set\n" );
		return -EFAULT;
	}
SIFS_PROFILE( "Private superblock info is set\n" );

	info = (struct sifs_sb_info*) file->f_dentry->d_sb->u.generic_sbp;
	memset( buffer, 0, 1040 );
	seqnr = get_seqnr( info );
SIFS_PROFILE( "Got sequence number\n" );

	if( ( error = sifs_getname( file->f_dentry, path ) ) < 0 ) {
		SIFS_DEBUG( "Rebuilding path from dentries failed\n" );
		return error;
	}
SIFS_PROFILE( "Rebuilding path from dentries succeeded, path = %s\n", path );

	if( ( len = str_encode( buffer + 14, path ) ) < 0 ) {
		SIFS_DEBUG( "Encoding path into request failed\n" );
		return len;
	}

	PUT_32BIT( buffer, seqnr );
	PUT_32BIT( buffer + 4, 2 + len );   // Packet length
	PUT_32BIT( buffer + 8, SIFS_READDIR );   // Opcode
	PUT_16BIT( buffer + 12, 1 );   // Number of strings
SIFS_PROFILE( "Request built successfully\n" );

	down( &(info->sem) );

	if( ( size = atomic_io( ssh_write, info->socket, buffer, 14 + len ) ) < 0 ) {
		SIFS_DEBUG( "Sending request to server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Request sent to server\n" );

	if( ( size = atomic_io( ssh_read, info->socket, buffer, 12 ) ) < 0 ) {
		SIFS_DEBUG( "Receiving reply header from server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Reply header received\n" );

	if( GET_32BIT( buffer ) != seqnr ) {
		SIFS_DEBUG( "Sequence number doesn't match\n" );
		status = -EPROTO;
		goto error;
	}
SIFS_PROFILE( "Sequence number matched\n" );

	if( ( error = GET_32BIT( buffer + 8 ) ) != SIFS_EOK ) {
		SIFS_DEBUG( "Readdir request failed\n" );
		switch( error ) {
			case SIFS_ENAMETOOLONG:
				status = -EBADF;
				break;
			case SIFS_ENOENT:
				status = -ENOENT;
				break;
			case SIFS_EACCES:
				status = -EACCES;
				break;
			case SIFS_ESERVER:
			default:
				status = -ENOMEM;
				break;
		}
		goto out;
	}
SIFS_PROFILE( "Readdir request succeeded\n" );

	packetlen = GET_32BIT( buffer + 4 );

	if( ( size = atomic_io( ssh_read, info->socket, buffer, 2 ) ) < 0 ) {
		SIFS_DEBUG( "Receiving reply payload from server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Reply payload length is ok\n" );

	bytes = 2;
	numstr = GET_16BIT( buffer );
SIFS_PROFILE( "file->f_pos = %ul\n", file->f_pos );

	while( pos < numstr && pos != file->f_pos && bytes < packetlen ) {
		if( ( size = atomic_io( ssh_read, info->socket, buffer, 2 ) ) < 0 ) {
			SIFS_DEBUG( "Receiving reply payload from server failed\n" );
			status = size;
			goto error;
		}
SIFS_PROFILE( "Received length of next string\n" );

		if( (len = GET_16BIT( buffer ) ) > SIFS_MAXNAMELEN ) {
			SIFS_DEBUG( "Protocol error - file name is longer than allowed\n" );
			goto error;
		}
SIFS_PROFILE( "String length is ok\n" );

		if( bytes + 2 + len > packetlen ) {
			SIFS_DEBUG( "Protocol error - payload is longer than packet length\n" );
			status = -EPROTO;
			goto error;
		}
SIFS_PROFILE( "String and packet length are ok\n" );

		if( ( size = atomic_io( ssh_read, info->socket, buffer + 2, len ) ) < 0 ) {
			SIFS_DEBUG( "Receiving reply payload from server failed\n" );
			status = size;
			goto error;
		}
SIFS_PROFILE( "Received name\n" );

		bytes += 2 + len;
		pos++;
SIFS_PROFILE( "Dropped string: %s\n", buffer + 2 );
	}

	while( pos < numstr  && bytes < packetlen ) {
		if( ( size = atomic_io( ssh_read, info->socket, buffer, 2 ) ) < 0 ) {
			SIFS_DEBUG( "Receiving reply payload from server failed\n" );
			status = size;
			goto error;
		}
SIFS_PROFILE( "Received length of next string\n" );

		if( (len = GET_16BIT( buffer ) ) > SIFS_MAXNAMELEN ) {
			SIFS_DEBUG( "Protocol error - file name is longer than allowed\n" );
			goto error;
		}
SIFS_PROFILE( "String length is ok\n" );

		if( bytes + 2 + len > packetlen ) {
			SIFS_DEBUG( "Protocol error - payload is longer than packet length\n" );
			status = -EPROTO;
			goto error;
		}
SIFS_PROFILE( "String and packet length are ok\n" );

		if( ( size = atomic_io( ssh_read, info->socket, buffer + 2, len ) ) < 0 ) {
			SIFS_DEBUG( "Receiving reply payload from server failed\n" );
			status = size;
			goto error;
		}
SIFS_PROFILE( "Received name\n" );

		if( pos == file->f_pos ) {
			struct qstr qname;

			switch( file->f_pos ) {
				case 0:
					if( filldir( dirent, ".", 1, 0, inode->i_ino, DT_DIR) < 0 ) {
						SIFS_DEBUG( "Inserting \".\" failed\n" );
						return 0;
					}
SIFS_PROFILE( "Inserted string: .\n" );
					break;
				case 1:
					if( filldir( dirent, "..", 2, 1, dentry->d_parent->d_inode->i_ino, DT_DIR ) < 0 ) {
						SIFS_DEBUG( "Inserting \"..\" failed\n" );
						return 0;
					}
SIFS_PROFILE( "Inserted string: ..\n" );
					break;
				default:
					if( ( error = str_decode( buffer, &filename ) ) < 0 ) {
						SIFS_DEBUG( "Decoding string from reply failed\n" );
						status = error;
						goto error;
					}
SIFS_PROFILE( "Decoding string succeeded\n" );

					ino = 0;
					qname.name = filename;
					qname.len = strlen( filename );

					if( ( ino = find_inode_number( file->f_dentry, &qname ) ) == 0 ) {
						ino = iunique( file->f_dentry->d_sb, 2 );
					}

					if( filldir( dirent, qname.name, qname.len, file->f_pos, ino, DT_UNKNOWN ) < 0 ) {
						SIFS_DEBUG( "Inserting string by calling filldir failed: %s\n", qname.name );
					}

SIFS_PROFILE( "Inserted string: %s\n", qname.name );
			}
			file->f_pos++;
		}

		bytes += size + 2;
		pos++;
	}

	status = 0;
	goto out;

error:
	up( &(info->sem) );
	sifs_reconnect( info );
	return status;

out:
	up( &(info->sem) );
	return status;
}



/*
 *	Open a file on the server and stores the resulting file descriptior into
 *	the private structure of the inode.
 */

int sifs_open( struct inode* inode, struct file* file )
{
	int len = 0;
	int error = 0;
	int status = 0;
	ssize_t size = 0;
	u_int32_t seqnr = 0;
	u_int8_t buffer[1048];
	char path[1024];
	struct sifs_sb_info* info = (struct sifs_sb_info*) inode->i_sb->u.generic_sbp;


SIFS_PROFILE( "Open file:\n" );

	if( info == NULL ) {
		SIFS_DEBUG( "Private superblock info not set\n" );
		return -EFAULT;
	}
SIFS_PROFILE( "Private superblock info is set\n" );

	memset( buffer, 0, 1048 );
	seqnr = get_seqnr( info );
SIFS_PROFILE( "Got sequence number\n" );

	if( ( error = sifs_getname( file->f_dentry, path ) ) < 0 ) {
		SIFS_DEBUG( "Rebuilding path from dentries failed\n" );
		return error;
	}
SIFS_PROFILE( "Rebuilding path from dentries succeeded, path = %s\n", path );

	if( ( len = str_encode( buffer + 22, path ) ) < 0 ) {
		SIFS_DEBUG( "Encoding path into request failed\n" );
		return len;
	}

	PUT_32BIT( buffer, seqnr );
	PUT_32BIT( buffer + 4, 10 + len );   // Packet length
	PUT_32BIT( buffer + 8, SIFS_OPEN );   // Opcode
	PUT_32BIT( buffer + 12, 0 );   // Zero field
	PUT_32BIT( buffer + 16, file->f_flags );   // Open flags
	PUT_16BIT( buffer + 20, 1 );   // Number of strings
SIFS_PROFILE( "Request built successfully\n" );

	down( &(info->sem) );

	if( ( size = atomic_io( ssh_write, info->socket, buffer, 22 + len ) ) < 0 ) {
		SIFS_DEBUG( "Sending request to server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Request sent to server\n" );

	if( ( size = atomic_io( ssh_read, info->socket, buffer, 12 ) ) < 0 ) {
		SIFS_DEBUG( "Receiving reply header from server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Reply header received\n" );

	if( GET_32BIT( buffer ) != seqnr ) {
		SIFS_DEBUG( "Sequence number doesn't match\n" );
		status = -EPROTO;
		goto error;
	}
SIFS_PROFILE( "Sequence number matched\n" );

	if( ( error = GET_32BIT( buffer + 8 ) ) != SIFS_EOK ) {
		SIFS_DEBUG( "Open request failed\n" );
		switch( error ) {
			case SIFS_ENAMETOOLONG:
				status = -ENAMETOOLONG;
				break;
			case SIFS_ENOENT:
				status = -ENOENT;
				break;
			case SIFS_ETEMP:
				status = -ETXTBSY;
				break;
			case SIFS_EACCES:
				status = -EACCES;
				break;
			case SIFS_ESERVER:
			default:
				status = -ENOMEM;
				break;
		}
		goto out;
	}
SIFS_PROFILE( "Open request succeeded\n" );

	if( GET_32BIT( buffer + 4 ) != 0 ) {
		SIFS_DEBUG( "Invalid payload length of reply\n" );
		status = -EPROTO;
		goto error;
	}
SIFS_PROFILE( "Reply payload length is ok\n" );

	status = 0;
	goto out;

error:
	up( &(info->sem) );
	sifs_reconnect( info );
	return status;

out:
	up( &(info->sem) );
	return status;
}



/*
 *	Closes the file identified by the file descriptor on the server.
 */

int sifs_release( struct inode* inode, struct file* file )
{
	int len = 0;
	int error = 0;
	int status = 0;
	ssize_t size = 0;
	u_int32_t seqnr = 0;
	u_int8_t buffer[1048];
	char path[1024];
	struct sifs_sb_info* info = (struct sifs_sb_info*) inode->i_sb->u.generic_sbp;


SIFS_PROFILE( "Release file:\n" );

	if( info == NULL ) {
		SIFS_DEBUG( "Private superblock info not set\n" );
		return -EFAULT;
	}
SIFS_PROFILE( "Private superblock info is set\n" );

	memset( buffer, 0, 1048 );
	seqnr = get_seqnr( info );
SIFS_PROFILE( "Got sequence number\n" );

	if( ( error = sifs_getname( file->f_dentry, path ) ) < 0 ) {
		SIFS_DEBUG( "Rebuilding path from dentries failed\n" );
		return error;
	}
SIFS_PROFILE( "Rebuilding path from dentries succeeded, path = %s\n", path );

	if( ( len = str_encode( buffer + 14, path ) ) < 0 ) {
		SIFS_DEBUG( "Encoding path into request failed\n" );
		return len;
	}

	PUT_32BIT( buffer, seqnr );
	PUT_32BIT( buffer + 4, 2 + len );   // Packet length
	PUT_32BIT( buffer + 8, SIFS_RELEASE );   // Opcode
	PUT_16BIT( buffer + 12, 1 );   // Number of strings
SIFS_PROFILE( "Request built successfully\n" );

	down( &(info->sem) );

	if( ( size = atomic_io( ssh_write, info->socket, buffer, 14 + len ) ) < 0 ) {
		SIFS_DEBUG( "Sending request to server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Request sent to server\n" );

	if( ( size = atomic_io( ssh_read, info->socket, buffer, 12 ) ) < 0 ) {
		SIFS_DEBUG( "Receiving reply header from server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Reply header received\n" );

	if( GET_32BIT( buffer ) != seqnr ) {
		SIFS_DEBUG( "Sequence number doesn't match\n" );
		status = -EPROTO;
		goto error;
	}
SIFS_PROFILE( "Sequence number matched\n" );

	if( ( error = GET_32BIT( buffer + 8 ) ) != SIFS_EOK ) {
		SIFS_DEBUG( "Release request failed\n" );
		switch( error ) {
			case SIFS_ENAMETOOLONG:
				status = -EBADF;
				break;
			case SIFS_ENOENT:
				status = -EBADF;
				break;
			case SIFS_EIO:
				status = -EIO;
				break;
			case SIFS_ESERVER:
			default:
				status = -EINTR;
				break;
		}
		goto out;
	}
SIFS_PROFILE( "Release request succeeded\n" );

	if( GET_32BIT( buffer + 4 ) != 0 ) {
		SIFS_DEBUG( "Invalid payload length of reply\n" );
		status = -EPROTO;
		goto error;
	}
SIFS_PROFILE( "Reply payload length is ok\n" );

	status = 0;
	goto out;

error:
	up( &(info->sem) );
	sifs_reconnect( info );
	return status;

out:
	up( &(info->sem) );
	return status;
}



/*
 *	Flush all unwritten data to the harddisk of the server
 */

int sifs_fsync( struct file* file, struct dentry* dentry, int datasync )
{
SIFS_PROFILE( "Sync cache with hard disk:\n" );
	return -EROFS;
}



/*
 *	Get and set locks for a (portion of a) file
 */

int sifs_lock( struct file* file, int cmd, struct file_lock* flock )
{
SIFS_PROFILE( "Lock file:\n" );
	return -ENOLCK;
}
