#include "sifs.h"



/*
 *	Implementation of inode related funtions
 */



struct inode_operations sifs_dir_inode_ops = {
	lookup:		sifs_lookup,
	create:		sifs_create,
	link:			sifs_link,
	unlink:		sifs_unlink,
	symlink:		sifs_symlink,
	mkdir:		sifs_mkdir,
	rmdir:			sifs_rmdir,
	mknod:		sifs_mknod,
	rename:		sifs_rename,
	permission:	sifs_permission,
	revalidate:	sifs_revalidate,
	setattr:		sifs_setattr,
	getattr:		sifs_getattr,
};

struct inode_operations sifs_file_inode_ops = {
	truncate:		sifs_truncate,
	permission:	sifs_permission,
	revalidate:	sifs_revalidate,
	setattr:		sifs_setattr,
	getattr:		sifs_getattr,
};

struct inode_operations sifs_symlink_inode_ops = {
	readlink:		sifs_readlink,
	follow_link:	sifs_follow_link,
	revalidate:	sifs_revalidate,
	setattr:		sifs_setattr,
	getattr:		sifs_getattr,
};



/*
 *	Generate a new inode
 */

struct inode* sifs_iget( struct super_block* sb, struct sifs_attr* attr )
{
	struct inode *result;


SIFS_PROFILE( "Create new inode:\n" );

	if( ( result = new_inode( sb ) ) == NULL ) {
		SIFS_DEBUG( "Unable to get new inode\n" );
		return NULL;
	}
SIFS_PROFILE( "Creating inode succeeded\n" );

	result->i_ino = attr->ia_ino;
	result->i_mode = attr->ia_mode;
	result->i_nlink = attr->ia_nlink;
	result->i_uid = attr->ia_uid;
	result->i_gid = attr->ia_gid;
	result->i_atime = attr->ia_atime;
	result->i_mtime = attr->ia_mtime;
	result->i_ctime = attr->ia_ctime;
	result->i_size = attr->ia_size;
	result->i_blksize = 8192;
	result->i_blocks = (unsigned long) attr->ia_size / 512;


	if( S_ISDIR( result->i_mode ) ) {
		result->i_op = &sifs_dir_inode_ops;
		result->i_fop = &sifs_dir_ops;
	} else if( S_ISLNK( result->i_mode ) ) {
		result->i_op = &sifs_symlink_inode_ops;
	} else {
		result->i_op = &sifs_file_inode_ops;
		result->i_fop = &sifs_file_ops;
		result->i_data.a_ops = &sifs_aops;
	}

	insert_inode_hash( result );
	return result;
}



/*
 *	Do a lookup for a path component and return the result
 */

struct dentry* sifs_lookup( struct inode* inode, struct dentry* dentry )
{
	int i = 0;
	int error = 0;
	struct iattr attr;
	struct sifs_attr sifs_attr;
	struct inode* new = NULL;


SIFS_PROFILE( "Lookup path name:\n" );

	memset( &attr, 0, sizeof( struct iattr ) );
	memset( &sifs_attr, 0, sizeof( struct sifs_attr ) );

	if( ( error = sifs_getattr( dentry, &attr ) ) < 0 ) {
		SIFS_DEBUG( "Path component doesn't exist, adding negative dentry\n" );
	}
SIFS_PROFILE( "getattr succeeded\n" );

	sifs_attr.ia_ino = iunique( inode->i_sb, 2 );
	sifs_attr.ia_nlink = 1;
	sifs_attr.ia_mode = attr.ia_mode;
	sifs_attr.ia_uid = attr.ia_uid;
	sifs_attr.ia_gid = attr.ia_gid;
	sifs_attr.ia_atime = attr.ia_atime;
	sifs_attr.ia_mtime = attr.ia_mtime;
	sifs_attr.ia_ctime = attr.ia_ctime;
	sifs_attr.ia_size = attr.ia_size;

	if( ( new = sifs_iget( inode->i_sb, &sifs_attr ) ) == NULL ) {
		SIFS_DEBUG( "Creating new inode failed\n" );
		error = -ENOMEM;   // Is this ok?
		goto error;
	}
SIFS_PROFILE( "iget succeeded\n" );

	d_add( dentry, new );
	sifs_renew_times( dentry );

error:
	return ERR_PTR( error );
}



/*
 *	Create a new file on the server
 */

int sifs_create( struct inode* inode, struct dentry* dentry, int mode )
{
SIFS_PROFILE( "Create new file:\n" );
	return -EROFS;
}



/*
 *	Make a hard link to an existing file on the server (provided the server os
 *	is supporting that)
 */

int sifs_link( struct dentry* old_denty, struct inode* inode, struct dentry* new_dentry )
{
SIFS_PROFILE( "Make hard link:\n" );
	return -EROFS;
}



/*
 *	Remove a file on the server
 */

int sifs_unlink( struct inode* inode, struct dentry* dentry )
{
SIFS_PROFILE( "Remove file:\n" );
	return -EROFS;
}



/*
 *	Make a symbolic link on the server
 */

int sifs_symlink( struct inode* inode, struct dentry* dentry, const char* symname )
{
SIFS_PROFILE( "Create symbolic link:\n" );
	return -EROFS;
}



/*
 *	Make a new directory on the server
 */

int sifs_mkdir( struct inode* inode, struct dentry* dentry, int mode )
{
SIFS_PROFILE( "Create new direcory:\n" );
	return -EROFS;
}



/*
 *	Remove an existing directory on the server
 */

int sifs_rmdir( struct inode* inode, struct dentry* dentry )
{
SIFS_PROFILE( "Remove direcory:\n" );
	return -EROFS;
}



/*
 *	Create a special file (pipe, socket, etc) on the server. This must be
 *	supported by the server os.
 */

int sifs_mknod( struct inode* inode, struct dentry* dentry, int mode, int dev )
{
SIFS_PROFILE( "Create special file:\n" );
	return -EROFS;
}



/*
 *	Rename a file on the server
 */

int sifs_rename( struct inode* old_inode, struct dentry* old_dentry, struct inode* new_inode, struct dentry* new_dentry )
{
SIFS_PROFILE( "Rename file:\n" );
	return -EROFS;
}



/*
 *	Truncate a file
 */

void sifs_truncate( struct inode* inode )
{
SIFS_PROFILE( "Truncate file:\n" );
}



/*
 *	Check, if the requested permissions through "mask" are compatible with
 *	that given from the inode.
 */

int sifs_permission( struct inode* inode, int mask )
{
SIFS_PROFILE( "Examine permissions:\n" );

	/*
	 * The Unix mode bits are ok, because of the uid and gid mapping done
	 * by sifs_getattr(). All files that don't belong to the user are mapped to
	 * the super user. In case of a file operation on the server by the super
	 * user, the operation may be denied because the super user only has
	 * the same rights as the user.
	 */

	return vfs_permission( inode, mask );
}



/*
 * Look, if the file or directory is still valid, i.e. it have to exist
 */

int sifs_revalidate( struct dentry* dentry )
{
	int error = 0;
	struct iattr attr;
	struct inode* inode = (struct inode*) dentry->d_inode;


SIFS_PROFILE( "revalidate inode:\n" );

	if( inode == NULL ) {
		SIFS_DEBUG( "No inode is assigned to the dentry\n" );
		return -EFAULT;
	}
SIFS_PROFILE( "Inode is set\n" );

	memset( &attr, 0, sizeof( struct iattr ) );

	if( ( error = sifs_getattr( dentry, &attr ) ) < 0 ) {
		SIFS_DEBUG( "Unable to retrieve file or directory attributes from server\n" );
		return error;
	}
SIFS_PROFILE( "getattr succeeded\n" );

	inode->i_mode = attr.ia_mode;
	inode->i_uid = attr.ia_uid;
	inode->i_gid = attr.ia_gid;
	inode->i_atime = attr.ia_atime;
	inode->i_mtime = attr.ia_mtime;
	inode->i_ctime = attr.ia_ctime;
	inode->i_size = attr.ia_size;

	return 0;
}



/*
 *	Update all attributes which have changed
 */

int sifs_setattr( struct dentry* dentry, struct iattr* attr )
{
SIFS_PROFILE( "Set new attributes:\n" );
	return -EROFS;
}



/*
 *	Read all inode attributes from the server
 */

int sifs_getattr( struct dentry* dentry, struct iattr* attr )
{
	int len = 0;
	int error = 0;
	int status = 0;
	ssize_t size = 0;
	u_int32_t seqnr = 0;
	u_int8_t buffer[1040];
	char path[1024];
	struct sifs_sb_info* info = (struct sifs_sb_info*) dentry->d_sb->u.generic_sbp;


SIFS_PROFILE( "Get inode attributes:\n" );

	if( info == NULL ) {
		SIFS_DEBUG( "Private superblock info is not set\n" );
		return -EFAULT;
	}
SIFS_PROFILE( "Private superblock info is set\n" );

	memset( buffer, 0, 1040 );
	seqnr = get_seqnr( info );
SIFS_PROFILE( "Got sequence number\n" );

	if( ( error = sifs_getname( dentry, path ) ) < 0 ) {
		SIFS_DEBUG( "Rebuilding path from dentries failed\n" );
		return error;
	}
SIFS_PROFILE( "Rebuilding path from dentries succeeded, path = %s\n", path );

	if( ( len = str_encode( buffer + 14, path ) ) < 0 ) {
		SIFS_DEBUG( "Encoding path into request failed\n" );
		return len;
	}

	PUT_32BIT( buffer, seqnr );
	PUT_32BIT( buffer + 4, 2 + len );   // Packet length
	PUT_32BIT( buffer + 8, SIFS_GETATTR );   // Opcode
	PUT_16BIT( buffer + 12, 1 );   // Number of strings
SIFS_PROFILE( "Request built successfully\n" );

	down( &(info->sem) );

	if( ( size = atomic_io( ssh_write, info->socket, buffer, 14 + len ) ) < 0 ) {
		SIFS_DEBUG( "Sending request to server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Request sent to server\n" );

	if( ( size = atomic_io( ssh_read, info->socket, buffer, 12 ) ) < 0 ) {
		SIFS_DEBUG( "Receiving reply header from server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Reply header received\n" );

	if( GET_32BIT( buffer ) != seqnr ) {
		SIFS_DEBUG( "Sequence number doesn't match\n" );
		status = -EPROTO;
		goto error;
	}
SIFS_PROFILE( "Sequence number matched\n" );

	if( ( error = GET_32BIT( buffer + 8 ) ) != SIFS_EOK ) {
		SIFS_DEBUG( "Getattr request failed\n" );
		switch( error ) {
			case SIFS_ENAMETOOLONG:
				status = -ENAMETOOLONG;
				break;
			case SIFS_ENOENT:
				status = -ENOENT;
				break;
			case SIFS_EACCES:
				status = -EACCES;
				break;
			case SIFS_ESERVER:
			default:
				status = -ENOMEM;
				break;
		}
		goto out;
	}
SIFS_PROFILE( "Getattr request succeeded\n" );

	if( GET_32BIT( buffer + 4 ) != 36 ) {
		SIFS_DEBUG( "Invalid payload length of reply\n" );
		status = -EPROTO;
		goto error;
	}
SIFS_PROFILE( "Reply payload length is ok\n" );

	if( ( size = atomic_io( ssh_read, info->socket, buffer, 36 ) ) < 0 ) {
		SIFS_DEBUG( "Receiving reply payload from server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Reply payload received sucessfully\n" );

	attr->ia_mode = GET_32BIT( buffer ) & 0170777;
	// number of links are dropped
	attr->ia_uid = GET_32BIT( buffer + 8 );
	attr->ia_gid = GET_32BIT( buffer + 12 );
	attr->ia_atime = GET_32BIT( buffer + 16 );
	attr->ia_mtime = GET_32BIT( buffer + 20 );
	attr->ia_ctime = GET_32BIT( buffer + 24 );
	attr->ia_size = GET_64BIT( buffer + 28 );

	if( attr->ia_uid == info->server.uid ) {   // uid mapping
		attr->ia_uid = info->client.uid;
	} else {
		attr->ia_uid = 0;
	}

	if( attr->ia_gid == info->server.uid ) {   // gid mapping
		attr->ia_gid = info->client.uid;
	} else {
		attr->ia_gid = 0;
	}

	status = 0;
	goto out;

error:
	up( &(info->sem) );
	sifs_reconnect( info );
	return status;

out:
	up( &(info->sem) );
	return status;
}



/*
 *	Retrieve link content (private funtion)
 */

int sifs_getlink( struct dentry* dentry, u_int8_t* linkbuf )
{
	int len = 0;
	int error = 0;
	int status = 0;
	ssize_t size = 0;
	u_int32_t seqnr = 0;
	char path[1024];
	char buffer[1040];
	struct sifs_sb_info* info = (struct sifs_sb_info*) dentry->d_sb->u.generic_sbp;


SIFS_PROFILE( "Retrieve link content:\n" );

	if( info == NULL ) {
		SIFS_DEBUG( "Private superblock info not set\n" );
		return -EFAULT;
	}
SIFS_PROFILE( "Private superblock info is set\n" );

	memset( buffer, 0, 1040 );
	seqnr = get_seqnr( info );
SIFS_PROFILE( "Got sequence number\n" );

	if( ( error = sifs_getname( dentry, path ) ) < 0 ) {
		SIFS_DEBUG( "Rebuilding path from dentries failed\n" );
		return error;
	}
SIFS_PROFILE( "Rebuilding path form dentries succeeded, path = %s\n", path );

	if( ( len = str_encode( buffer + 14, path ) ) < 0 ) {
		SIFS_DEBUG( "Encoding path into request failed\n" );
		return len;
	}

	PUT_32BIT( buffer, seqnr );
	PUT_32BIT( buffer + 4, 2 + len );   // Packet length
	PUT_32BIT( buffer + 8, SIFS_READLINK );   // Opcode
	PUT_16BIT( buffer + 12, 1 );   // Number of strings
SIFS_PROFILE( "Request built successfully\n" );

	down( &(info->sem) );

	if( ( size = atomic_io( ssh_write, info->socket, buffer, 14 + len ) ) < 0 ) {
		SIFS_DEBUG( "Sending request to server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Request sent to server\n" );

	if( ( size = atomic_io( ssh_read, info->socket, buffer, 12 ) ) < 0 ) {
		SIFS_DEBUG( "Receiving reply header from server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Reply header received\n" );

	if( GET_32BIT( buffer ) != seqnr ) {
		SIFS_DEBUG( "Sequence number doesn't match\n" );
		status = -EPROTO;
		goto error;
	}
SIFS_PROFILE( "Sequence number matched\n" );

	if( ( error = GET_32BIT( buffer + 8 ) ) != SIFS_EOK ) {
		SIFS_DEBUG( "Readlink request failed\n" );
		switch( error ) {
			case SIFS_ENAMETOOLONG:
				status = -ENAMETOOLONG;
				break;
			case SIFS_ENOENT:
				status = -ENOENT;
				break;
			case SIFS_EINVAL:
				status = -EINVAL;
				break;
			case SIFS_EACCES:
				status = -EACCES;
				break;
			case SIFS_EIO:
				status = -EIO;
				break;
			case SIFS_ESERVER:
			default:
				status = -ENOMEM;
				break;
		}
		goto out;
	}
SIFS_PROFILE( "Readlink request succeeded\n" );

	if( ( len = GET_32BIT( buffer + 4 ) ) > 1028 ) {
		SIFS_DEBUG( "Invalid payload length of reply\n" );
		status = -EPROTO;
		goto error;
	}
SIFS_PROFILE( "Reply payload length is ok\n" );
SIFS_PROFILE( "Reply length = %d\n", len );

	if( ( size = atomic_io( ssh_read, info->socket, buffer, 4 ) ) < 0 ) {
		SIFS_DEBUG( "Receiving reply payload from server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Number of strings and string length received\n" );

	if( GET_16BIT( buffer ) != 1 ) {
		SIFS_DEBUG( "Invalid string number of reply\n" );
		status = -EPROTO;
		goto error;
	}
SIFS_PROFILE( "Number of strings is ok\n" );

	if( ( size = GET_16BIT( buffer + 2 ) ) > 1024 ) {
		SIFS_DEBUG( "Invalid length of reply string\n" );
		status = -EPROTO;
		goto error;
	}
SIFS_PROFILE( "Length of reply string is ok\n" );

	if( size + 4 != len ) {
		SIFS_DEBUG( "Packet length field and payload length are not the same\n" );
		status = -EPROTO;
		goto error;
	}
SIFS_PROFILE( "Packet length field and payload length are equal\n" );

	if( ( size = atomic_io( ssh_read, info->socket, linkbuf, size ) ) < 0 ) {
		SIFS_DEBUG( "Receiving reply payload from server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Reply payload received\n" );

	status = 0;
	goto out;

error:
	up( &(info->sem) );
	sifs_reconnect( info );
	return status;

out:
	up( &(info->sem) );
	return status;
}



/*
 *	Read the contents of the link (where it points to)
 */

int sifs_readlink( struct dentry* dentry, char* buffer, int buflen )
{
	int result = 0;
	u_int8_t linkbuf[1024];


SIFS_PROFILE( "Read link:\n" );

	if( ( result = sifs_getlink( dentry, linkbuf ) < 0 ) ) {
		SIFS_DEBUG( "Retrieving link content failed\n" );
		goto error;
	}
SIFS_PROFILE( "getlink succeeded\n" );

	result = vfs_readlink( dentry, buffer, buflen, linkbuf );

error:
	return result;
}



/*
 *	Get the inode where the link points to
 */

int sifs_follow_link( struct dentry* dentry, struct nameidata* nd )
{
	int result = 0;
	u_int8_t linkbuf[1024];


SIFS_PROFILE( "Follow link:\n" );

	if( ( result = sifs_getlink( dentry, linkbuf ) < 0 ) ) {
		SIFS_DEBUG( "Retrieving link content failed\n" );
		goto error;
	}
SIFS_PROFILE( "getlink succeeded\n" );

	result = vfs_follow_link( nd, linkbuf );

error:
	return result;
}
