#ifndef SIFS_H
#define SIFS_H



#include "ssh.h"
#include "sifs-codes.h"



/*
 *	Some useful constants
 */

#define SIFS_SUPER_MAGIC 0xAAEF2002
#define SIFS_MAXPATHLEN 1024
#define SIFS_MAXNAMELEN 256



/*
 *	Debugging and profiling macros
 */

#ifdef SIFS_DEBUG
#define SIFS_DEBUG(x...) printk( KERN_NOTICE __FUNCTION__ ": " x)
#else
#define SIFS_DEBUG(x...) do { ; } while(0)
#endif

#ifdef SIFS_PROFILE
#define SIFS_PROFILE(x...) printk( KERN_NOTICE __FUNCTION__ ": " x)
#else
#define SIFS_PROFILE(x...) do { ; } while(0)
#endif



/*
 * Internal used structures for data handling
 */

struct ids {
	uid_t uid;
	gid_t gid;
};

struct sifs_sb_info {
	char serverpath[1024];
	char username[256];
	char password[256];
	struct sshsocket* socket;
	struct sockaddr_in address;
	struct semaphore sem;
	struct ids server;
	struct ids client;
	u_int32_t seqnr;
};

struct sifs_attr
{
	unsigned long ia_ino;
	umode_t ia_mode;
	nlink_t ia_nlink;
	uid_t ia_uid;
	gid_t ia_gid;
	time_t ia_atime;
	time_t ia_mtime;
	time_t ia_ctime;
	loff_t ia_size;
};


/*
 *	Super block related
 */

static struct super_block* sifs_read_super( struct super_block*, void*, int );
void sifs_put_super( struct super_block* );
int sifs_statfs( struct super_block*, struct statfs* );
int sifs_mount( struct sifs_sb_info* );
int sifs_unmount( struct sifs_sb_info* );


/*
 *	File related
 */

int sifs_readdir( struct file*, void*, filldir_t );
int sifs_open( struct inode*, struct file* );
int sifs_release( struct inode*, struct file* );
int sifs_fsync( struct file*, struct dentry*, int );
int sifs_lock( struct file*, int, struct file_lock* );




/*
 *	Inode related
 */

struct inode* sifs_iget( struct super_block*, struct sifs_attr* );
struct dentry *sifs_lookup( struct inode*, struct dentry* );
int sifs_create( struct inode* , struct dentry*, int );
int sifs_link( struct dentry*, struct inode*, struct dentry* );
int sifs_unlink( struct inode*, struct dentry* );
int sifs_symlink( struct inode*, struct dentry*, const char* );
int sifs_mkdir( struct inode*, struct dentry*, int );
int sifs_rmdir( struct inode*, struct dentry* );
int sifs_mknod( struct inode*, struct dentry*, int, int );
int sifs_rename( struct inode*, struct dentry*, struct inode*, struct dentry* );
int sifs_readlink( struct dentry*, char*, int );
int sifs_follow_link( struct dentry*, struct nameidata* );
void sifs_truncate( struct inode* );
int sifs_permission( struct inode*, int );
int sifs_revalidate( struct dentry* );
int sifs_setattr( struct dentry*, struct iattr* );
int sifs_getattr( struct dentry*, struct iattr* );


/*
 *	Address space related
 */

int sifs_readpage( struct file*, struct page* );
int sifs_writepage( struct page* );
int sifs_prepare_write( struct file*, struct page*, unsigned int, unsigned int );
int sifs_commit_write( struct file*, struct page*, unsigned int, unsigned int );



/*
 *	SIFS string related functions
 */

int str_encode( u_int8_t*, char* );
int str_decode( u_int8_t*, char** );



/*
 *	Supplementary functions
 */

int sifs_parse_options( struct sifs_sb_info*, char* );
int sifs_reconnect( struct sifs_sb_info* );
int sifs_getname( struct dentry*, char* );
void sifs_renew_times( struct dentry* );
inline u_int32_t get_seqnr( struct sifs_sb_info* );
ssize_t atomic_io( ssize_t (*fcn)( struct sshsocket*, void*, size_t ), struct sshsocket*, void*, size_t );



/*
 *	Structs with a lot of function pointers
 */

extern struct inode_operations sifs_dir_inode_ops;
extern struct inode_operations sifs_file_inode_ops;
extern struct inode_operations sifs_symlink_inode_ops;
extern struct file_operations sifs_dir_ops;
extern struct file_operations sifs_file_ops;
extern struct address_space_operations sifs_aops;

#endif
