/*
 *	Layer to the ssh implementation
 *	Currently only used to implement the packet protocol for the client sifsd
 */



#include "ssh.h"



/*
 * Read or write a chunk of data at once
 */

ssize_t ssh_io( ssize_t (*fcn)( struct socket*, void*, size_t ), struct socket* sock, void* buf, size_t count )
{
	ssize_t cnt = 0;
	size_t bytes = 0;


	while( bytes < count ) {
		if( ( cnt = (fcn)( sock, buf + bytes, count - bytes ) ) <= 0 ) {
			if( cnt == 0 ) {
				return -1;
			}
			return cnt;
		}

		bytes += cnt;
	}

	return bytes;
}



/*
 *	Create a communication endpoint (a socket) from kernelspace to sifsd
 */

struct sshsocket* ssh_create( int domain, int type, int protocol  )
{
	struct sshsocket* sock;


	if( ( sock = (struct sshsocket*) kmalloc( sizeof( struct sshsocket ), GFP_KERNEL ) ) == NULL ) {
		return NULL;
	}

	memset( sock, 0, sizeof( struct sshsocket ) );

	if( ( sock->socket = socket_create( domain, type, protocol ) ) < 0 ) {
		kfree( sock );
		return NULL;
	}

	return sock;
}



/*
 *	Open a connection from kernelspace to the server via sifsd
  */

int ssh_connect( struct sshsocket* sock, struct sockaddr* address, int addrlen, char* user, char* passwd )
{
	int error = 0;
	int sizeuser = sizeof( user ) + 1;
	int sizepasswd = sizeof( passwd ) + 1;
	u_int8_t connectmsg[523];


	if( sock == NULL || sizeuser + sizepasswd > 512 ) {
		return -EINVAL;
	}

	memset( connectmsg, 0, 523 );

	if( ( error = socket_connect( sock->socket, address, addrlen ) ) < 0 ) {
		printk( "ssh_connect(): Unable to connect to client deamon\n" );
		return error;
	}

	PUT_32BIT( connectmsg, 6 + sizeuser + sizepasswd );   // num_strings + size(userstring) + userstring + size(passwdstring) + passwdstring
	connectmsg[4] = 0;   // msgtype: connect
	PUT_16BIT( connectmsg + 5, 2 );   // two strings

	PUT_16BIT( connectmsg + 7, sizeuser );   // size of user string
	memcpy( connectmsg + 9, user, sizeuser );

	PUT_16BIT( connectmsg + 9 + sizeuser, sizepasswd );   // size of password string
	memcpy( connectmsg + 11 + sizeuser, passwd, sizepasswd );

	if( ( error = ssh_io( socket_write, sock->socket, connectmsg, 11+sizeuser+sizepasswd ) ) <= 0 ) {
		printk( "ssh_connect(): Unable to send connect message to client daemon\n" );
		return error;
	}

	return 0;
}



/*
 *	Send a disconnect and terminate signal to sifsd
 */

int ssh_disconnect( struct sshsocket* sock )
{
	int error = 0;
	u_int8_t header[5];


	if( sock == NULL ) {
		return -EINVAL;
	}

	PUT_32BIT( header, 5 );
	header[4] = 1;   // Disconnect

	if( ( error = ssh_io( socket_write, sock->socket, header, 5 ) ) != 5 ) {
		return error;
	}

	return 0;
}



/*
 *	Close a connection from kernelspace to sifsd
 */

int ssh_close( struct sshsocket* sock )
{
	int error = 0;


	if( sock == NULL ) {
		return -EINVAL;
	}

	if( ( error = socket_close( sock->socket ) ) < 0 ) {
		return error;
	}

	kfree( sock );
	return 0;
}



/*
 *	Receive message from the server via sifsd
 */

ssize_t ssh_read( struct sshsocket* sock, void* buffer, size_t count )
{
	ssize_t size = 0;
	u_int8_t header[5];


	if( sock == NULL ) {
		return -EINVAL;
	}

	// No more data from the previous packet is available, read a new one

	if( sock->avail == 0 ) {
		if( ( size = ssh_io( socket_read, sock->socket, header, 5 ) ) < 0 ) {
			return size;
		}

		sock->avail = GET_32BIT( header );

		if( header[4] != 2 ) {   // Only data packets are allowed
			return -EPROTO;
		}
	}

	size = sock->avail;
	if( count < size ) {
		size = count;
	}

	if( ( size = ssh_io( socket_read, sock->socket, buffer, size ) ) < 0 ) {
		return size;
	}

	sock->avail -= size;
	return size;
}



/*
 *	Send message from kernelspace to the server via sifsd
 */

ssize_t ssh_write( struct sshsocket* sock, void* buffer, size_t count )
{
	ssize_t size = 0;
	u_int8_t header[5];


	if( sock == NULL ) {
		return -EINVAL;
	}

	PUT_32BIT( header, count );
	header[4] = 2;   // data packet

	if( ( size = ssh_io( socket_write, sock->socket, header, 5 ) ) < 0 ) {
		return size;
	}

	return ssh_io( socket_write, sock->socket, buffer, count );
}
