/*
 *	Implementation of all super block related functions
 *	and the module stuff
 */



#include "sifs.h"



static struct super_operations sifs_sops = {
	statfs:			sifs_statfs,
	put_super:		sifs_put_super,
	put_inode:		force_delete,
};



int sifs_mount( struct sifs_sb_info* info )
{
	int len = 0;
	int status = 0;
	ssize_t size = 0;
	u_int32_t seqnr = 0;
	u_int8_t buffer[1044];


SIFS_PROFILE( "Mount remote filesystem:\n" );

	if( info == NULL ) {
		SIFS_DEBUG( "Private superblock info is not set\n" );
		return -EINVAL;
	}
SIFS_PROFILE( "Private superblock info is set\n" );

	memset( buffer, 0, 1044 );
	seqnr = get_seqnr( info );
SIFS_PROFILE( "Got sequence number\n" );

	if( ( len = str_encode( buffer + 18, info->serverpath ) ) < 0 ) {
		SIFS_DEBUG( "Encoding server path into request failed\n" );
		return len;
	}

	PUT_32BIT( buffer, seqnr );
	PUT_32BIT( buffer + 4, 6 + len );   // Packet length
	PUT_32BIT( buffer + 8, SIFS_MOUNT );   // Opcode
	PUT_32BIT( buffer + 12, 1 );   // Version
	PUT_16BIT( buffer + 16, 1 );   // Number of strings
SIFS_PROFILE( "Request built successfully\n" );

	down( &(info->sem) );

	if( ( size = atomic_io( ssh_write, info->socket, buffer, 18 + len ) ) < 0 ) {
		SIFS_DEBUG( "Sending request to server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Request sent to server\n" );

	if( ( size = atomic_io( ssh_read, info->socket, buffer, 12 ) ) < 0 ) {
		SIFS_DEBUG( "Receiving reply header from server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Reply header received\n" );

	if( GET_32BIT( buffer ) != seqnr ) {
		SIFS_DEBUG( "Sequence number doesn't match\n" );
		status = -EPROTO;
		goto error;
	}
SIFS_PROFILE( "Sequence number matched\n" );

	if( ( status = GET_32BIT( buffer + 8 ) ) != SIFS_EOK ) {
		SIFS_DEBUG( "Mount request failed\n" );
		goto error;
	}
SIFS_PROFILE( "Mount request succeeded\n" );

	if( GET_32BIT( buffer + 4 ) != 8 ) {
		SIFS_DEBUG( "Invalid payload length of reply\n" );
		status = -EPROTO;
		goto error;
	}
SIFS_PROFILE( "Reply payload length is ok\n" );

	if( ( size = atomic_io( ssh_read, info->socket, buffer, 8 ) ) < 0 ) {
		SIFS_DEBUG( "Receiving reply payload from server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Reply payload received sucessfully\n" );

	info->server.uid = GET_32BIT( buffer );
	info->server.gid = GET_32BIT( buffer + 4 );
	status = 0;

error:
	up( &(info->sem) );
	return status;
}



int sifs_unmount( struct sifs_sb_info* info )
{
	int len = 0;
	int status = 0;
	ssize_t size = 0;
	u_int32_t seqnr = 0;
	u_int8_t buffer[1040];


SIFS_PROFILE( "Unmount remote filesystem:\n" );

	if( info == NULL ) {
		SIFS_DEBUG( "Private superblock info not set\n" );
		return -EINVAL;
	}
SIFS_PROFILE( "Private superblock info is set\n" );

	memset( buffer, 0, 1040 );
	seqnr = get_seqnr( info );
SIFS_PROFILE( "Got sequence number\n" );

	if( ( len = str_encode( buffer + 14, info->serverpath ) ) < 0 ) {
		SIFS_DEBUG( "Encoding server path into request failed\n" );
		return len;
	}

	PUT_32BIT( buffer, seqnr );
	PUT_32BIT( buffer + 4, 2 + len );   // Packet length
	PUT_32BIT( buffer + 8, SIFS_UNMOUNT );   // Opcode
	PUT_16BIT( buffer + 12, 1 );   // Number of strings
SIFS_PROFILE( "Request built successfully\n" );

	down( &(info->sem) );

	if( ( size = atomic_io( ssh_write, info->socket, buffer, 14 + len ) ) < 0 ) {
		SIFS_DEBUG( "Sending request to server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Request sent to server\n" );

	if( ( size = atomic_io( ssh_read, info->socket, buffer, 12 ) ) < 0 ) {
		SIFS_DEBUG( "Receiving reply header from server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Reply header received\n" );

	if( GET_32BIT( buffer ) != seqnr ) {
		SIFS_DEBUG( "Sequence number doesn't match\n" );
		status = -EPROTO;
		goto error;
	}
SIFS_PROFILE( "Sequence number matched\n" );

	if( ( status = GET_32BIT( buffer + 8 ) ) != SIFS_EOK ) {
		SIFS_DEBUG( "Unmount request failed\n" );
		goto error;
	}
SIFS_PROFILE( "Unmount request succeeded\n" );

	if( GET_32BIT( buffer + 4 ) != 0 ) {
		SIFS_DEBUG( "Invalid payload length of reply\n" );
		status = -EPROTO;
		goto error;
	}
SIFS_PROFILE( "Reply payload length is ok\n" );

	status = 0;

error:
	up( &(info->sem) );
	return status;
}



/*
 *	Return status information about the file system
 */

int sifs_statfs( struct super_block* sb, struct statfs* buf )
{
	int len = 0;
	int error = 0;
	int status = 0;
	ssize_t size = 0;
	u_int32_t seqnr = 0;
	u_int8_t buffer[1040];
	struct sifs_sb_info* info = (struct sifs_sb_info*) sb->u.generic_sbp;


SIFS_PROFILE( "Get filesystem information:\n" );

	if( info == NULL ) {
		SIFS_DEBUG( "Private superblock info is not set\n" );
		return -EFAULT;
	}
SIFS_PROFILE( "Private superblock info is set\n" );

	memset( buffer, 0, 1040 );
	seqnr = get_seqnr( info );
SIFS_PROFILE( "Got sequence number\n" );

	if( ( len = str_encode( buffer + 14, "." ) ) < 0 ) {
		SIFS_DEBUG( "Encoding current directory into request failed\n" );
		return len;
	}

	PUT_32BIT( buffer, seqnr );
	PUT_32BIT( buffer + 4, 2 + len );   // Packet length
	PUT_32BIT( buffer + 8, SIFS_STATFS );   // Opcode
	PUT_16BIT( buffer + 12, 1 );   // Number of strings
SIFS_PROFILE( "Request built successfully\n" );

	down( &(info->sem) );

	if( ( size = atomic_io( ssh_write, info->socket, buffer, 14 + len ) ) < 0 ) {
		SIFS_DEBUG( "Sending request to server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Request sent to server\n" );

	if( ( size = atomic_io( ssh_read, info->socket, buffer, 12 ) ) < 0 ) {
		SIFS_DEBUG( "Receiving request header from server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Reply header received\n" );

	if( GET_32BIT( buffer ) != seqnr ) {
		SIFS_DEBUG( "Sequence number doesn't match\n" );
		status = -EPROTO;
		goto error;
	}
SIFS_PROFILE( "Sequence number matched\n" );

	if( ( error = GET_32BIT( buffer + 8 ) ) != SIFS_EOK ) {
		SIFS_DEBUG( "Statfs request failed\n" );
		switch( error ) {
			case SIFS_ENAMETOOLONG:
				status = -ENAMETOOLONG;
				break;
			case SIFS_ENOENT:
				status = -ENOENT;
				break;
			case SIFS_EACCES:
				status = -EACCES;
				break;
			case SIFS_EIO:
				status = -EIO;
				break;
			case SIFS_ENOTSUP:
				status = -ENOSYS;
				break;
			case SIFS_ESERVER:
			default:
				status = -ENOMEM;
				break;
		}
		goto out;
	}
SIFS_PROFILE( "Statfs request succeeded\n" );

	if( GET_32BIT( buffer + 4 ) != 28 ) {
		SIFS_DEBUG( "Invalid payload length of reply\n" );
		status = -EPROTO;
		goto error;
	}
SIFS_PROFILE( "Reply payload length is ok\n" );

	if( ( size = atomic_io( ssh_read, info->socket, buffer, 28 ) ) < 0 ) {
		SIFS_DEBUG( "Receiving request payload from server failed\n" );
		status = size;
		goto error;
	}
SIFS_PROFILE( "Reply payload received sucessfully\n" );

	buf->f_type = SIFS_SUPER_MAGIC;
	buf->f_bsize = 8192;
	buf->f_blocks = GET_32BIT( buffer + 4 );
	buf->f_bfree = GET_32BIT( buffer + 8 );
	buf->f_bavail = GET_32BIT( buffer + 12 );
	buf->f_files = GET_32BIT( buffer + 16 );
	buf->f_ffree = GET_32BIT( buffer + 20 );
	buf->f_namelen = GET_32BIT( buffer + 24 );

	status = 0;
	goto out;

error:
	up( &(info->sem) );
	sifs_reconnect( info );
	return status;

out:
	up( &(info->sem) );
	return status;
}



/*
 *	Connect to server and create root inode
 */

static struct super_block* sifs_read_super( struct super_block* sb, void* options, int silent )
{
	struct iattr attr;
	struct sifs_attr sifs_attr;
	struct dentry dentry;
	struct inode* root = NULL;
	struct sifs_sb_info* info = NULL;


SIFS_PROFILE( "Connect to server and mount filesystem:\n" );

	if( ( info = (struct sifs_sb_info*) kmalloc( sizeof( struct sifs_sb_info ), GFP_KERNEL ) ) == NULL ) {
		SIFS_DEBUG( "Out of memory\n" );
		return NULL;
	}
SIFS_PROFILE( "Allocating memory was successful\n" );

	memset( info, 0, sizeof( struct sifs_sb_info ) );
	memset( &sifs_attr, 0, sizeof( struct sifs_attr ) );
	memset( &attr, 0, sizeof( struct iattr ) );
	memset( &dentry, 0, sizeof( struct dentry ) );

	if( sifs_parse_options( info, (char*) options ) < 0 ) {
		if( !silent ) { SIFS_DEBUG( "Invalid options\n" ); }
		goto error;
	}
SIFS_PROFILE( "Mount options are ok\n" );

	init_MUTEX(&info->sem);
	sb->u.generic_sbp = info;
	sb->s_blocksize = 8192;
	sb->s_blocksize_bits = 13;
	sb->s_magic = SIFS_SUPER_MAGIC;
	sb->s_op = &sifs_sops;
	sb->s_flags = 0;

	if( sifs_reconnect( info ) < 0 ) {
		if( !silent ) { SIFS_DEBUG( "Connect to server failed\n" ); }
		goto error;
	}
SIFS_PROFILE( "Connect to server succeeded\n" );

	dentry.d_parent = &dentry;   // Root points to itself
	dentry.d_sb = sb;

	if( sifs_getattr( &dentry, &attr ) < 0 ) {
		SIFS_DEBUG( "Unable to get root directory attributes\n" );
		ssh_disconnect( info->socket );
		ssh_close( info->socket );
		goto error;
	}
SIFS_PROFILE( "getattr succeeded\n" );

	sifs_attr.ia_ino = 2;   // Root inode number
	sifs_attr.ia_ino = iunique( sb, 2 );
	sifs_attr.ia_nlink = 1;
	sifs_attr.ia_mode = attr.ia_mode;
	sifs_attr.ia_uid = attr.ia_uid;
	sifs_attr.ia_gid = attr.ia_gid;
	sifs_attr.ia_atime = attr.ia_atime;
	sifs_attr.ia_mtime = CURRENT_TIME;
	sifs_attr.ia_ctime = attr.ia_ctime;
	sifs_attr.ia_size = attr.ia_size;

	if( ( root = sifs_iget( sb, &sifs_attr ) ) == NULL ) {
		SIFS_DEBUG( "Unable to get root inode\n" );
		goto error;
	}
SIFS_PROFILE( "Got root inode\n" );

	if( ( sb->s_root = d_alloc_root( root ) ) == NULL ) {
		SIFS_DEBUG( "Unable to allocate dentry for root inode\n" );
		iput( root );
		goto error;
	}
SIFS_PROFILE( "Allocating dentry for root inode succeeded\n" );

	return sb;

error:
	memset( info, 0, sizeof( struct sifs_sb_info ) );
	kfree( info );

	return NULL;
}



/*
 *	Disconnect from server and free all resources
 */

void sifs_put_super( struct super_block* sb )
{
	struct sifs_sb_info* info = (struct sifs_sb_info*) sb->u.generic_sbp;


SIFS_PROFILE( "Unmount and disconnect from server:\n" );

	if( info == NULL ) {
		SIFS_DEBUG( "Private superblock info is not set\n" );
		return;
	}
SIFS_PROFILE( "Private superblock info is set\n" );

	if( sifs_unmount( info ) < 0 ) {
		SIFS_DEBUG( "Unable to unmount remote filesystem\n" );
	}
SIFS_PROFILE( "Unmount request succeeded\n" );

	ssh_disconnect( info->socket );
	ssh_close( info->socket );
	kfree( info );

}



/*
 *	Declare struct sifs_fs_type
 */

static DECLARE_FSTYPE( sifs_fs_type, "sifs", sifs_read_super, 0 );



/*
 * We have to register the file system (make the functionality known to the
 kernel)
 * when the module is loaded or at boot time.
 */

static int __init init_sifs_fs( void )
{
SIFS_PROFILE( "Registering sifs filesystem:\n" );
	return register_filesystem( &sifs_fs_type );
}



/*
 * Modules can be removed from the kernel, if they are not used any more. Then
 * we have to remove the information, that there is a file system called sifs
 * currently available (as long as not loaded again).
 */

static void __exit exit_sifs_fs( void )
{
SIFS_PROFILE( "Unregistering sifs filesystem:\n" );
	unregister_filesystem( &sifs_fs_type );
}



#ifdef MODULE

EXPORT_NO_SYMBOLS;


/*
 * The kernel have to know which functions are responsible for registering and
 * deregistering the module.
 */

module_init( init_sifs_fs )
module_exit( exit_sifs_fs )

#endif
