/*
 *	Supplementary functions for sifs filesystem driver
 */



#include "sifs.h"



/*
 * Convert ip string to number
 * Originally taken from ftpfs
 */

unsigned long sifs_inet_addr( char* ip )
{
	unsigned long res = 0;
	int i, no = 0, np = 0;

	for(i = 0; i < strlen(ip); i++){
		if(((ip[i]<'0')||(ip[i]>'9'))&&(ip[i]!='.')) return -1;
		if(ip[i] == '.'){
			if(++np >  3)
				return -1;
			if((no < 0)||(no > 255))
				return -1;
			res = (res >> 8) + (no << 24);
			no = 0;
		}else
			no = no*10 + ip[i] - '0';
	}
	if((no < 0)||(no > 255)) return -1;
	res = (res >> 8) + (no << 24);
	if(np != 3) return -1;

	return res;
}



/*
 *	Parse info from mount command
 * - ip address
 * - port
 * - local uid
 * - local gid
 * - remote username
 * - remote password
 */

int sifs_parse_options( struct sifs_sb_info* info, char* options )
{
	int i = 0;
	int idx = 0;
	int len = 0;
	char* opts[7] = { NULL };


SIFS_PROFILE( "Parse mount options\n" );

	if( info == NULL || options == NULL ) {
		SIFS_DEBUG( "Private superblock info or options not set\n" );
		return -EINVAL;
	}

	opts[idx] = options;

	while( options[i] != '\0' && idx < 6 ) {
		if( options[i] == ',' ) {
			options[i] = '\0';
			opts[++idx] = options + i + 1;
		}
		i++;
	}

	for( i = 0; i <= idx; i++ )
	{
		if( strncmp( opts[i], "ip=", 3 ) == 0 ) {
			if( strlen( opts[i] + 3 ) > 15 ) {
			    goto invalid_opts;
			}
			info->address.sin_family = PF_INET;
			info->address.sin_addr.s_addr = sifs_inet_addr( opts[i] + 3 );
		} else if( strncmp( opts[i], "port=", 5 ) == 0 ) {
			if( strlen( opts[i] + 5 ) > 5 ) {
			    goto invalid_opts;
			}
			info->address.sin_port = htons( simple_strtoul( opts[i] + 5, NULL, 0 ) );
		} else if( strncmp( opts[i], "serverpath=", 11 ) == 0 ) {
			len = strlen( opts[i] + 11 );
			if( len > 1024 ) {
			    goto invalid_opts;
			}
			memcpy( info->serverpath, opts[i] + 11, len );
			info->serverpath[len] = '\0';
		} else if( strncmp( opts[i], "uid=", 4 ) == 0 ) {
			if( strlen( opts[i] + 4 ) > 5 ) {
			    goto invalid_opts;
			}
			info->client.uid = simple_strtoul( opts[i] + 4, NULL, 0 );
		} else if( strncmp( opts[i], "gid=", 4 ) == 0 ) {
			if( strlen( opts[i] + 4 ) > 5 ) {
			    goto invalid_opts;
			}
			info->client.gid = simple_strtoul( opts[i] + 4, NULL, 0 );
		} else if( strncmp( opts[i], "username=", 9 ) == 0 ) {
			len = strlen( opts[i] + 9 );
			if( len > 255 ) {
			    goto invalid_opts;
			}
			memcpy( info->username, opts[i] + 9, len );
			info->username[len] = '\0';
		} else if( strncmp( opts[i], "password=", 9 ) == 0 ) {
			len = strlen( opts[i] + 9 );
			if( len > 255 ) {
			    goto invalid_opts;
			}
			memcpy( info->password, opts[i] + 9, len );
			info->password[len] = '\0';
		}
	}

	return 0;

invalid_opts:
	return -EINVAL;
}



/*
 * Increase and return the sequence number in a secure way
 */

inline u_int32_t get_seqnr( struct sifs_sb_info* info )
{
	u_int32_t seqnr;

	down( &(info->sem) );
	seqnr = info->seqnr++;
	up( &(info->sem) );

	return seqnr;
}



/*
 * Read or write a chunk of data at once
 */

ssize_t atomic_io( ssize_t (*fcn)( struct sshsocket*, void*, size_t ), struct sshsocket* sock, void* buf, size_t count )
{
	ssize_t cnt = 0;
	size_t bytes = 0;


	while( bytes < count ) {
		if( ( cnt = (fcn)( sock, buf + bytes, count - bytes ) ) <= 0 ) {
			if( cnt == 0 ) {
				return -EIO;
			}
			return cnt;
		}

		bytes += cnt;
	}

	return bytes;
}



/*
 *	Close an old connection and establish a new one
 */

int sifs_reconnect( struct sifs_sb_info* info )
{
SIFS_PROFILE( "(Re)connect to server\n" );

	if( info->socket != NULL ) {
		ssh_close( info->socket );
	}

	if( ( info->socket = ssh_create( PF_INET, SOCK_STREAM, 0 ) ) == NULL ) {
		SIFS_DEBUG( "Unable to create communication socket\n" );
		goto error;
	}

	if( ssh_connect( info->socket, (struct sockaddr*) &(info->address), sizeof( struct sockaddr ), info->username, info->password ) < 0 ) {
		SIFS_DEBUG( "Unable to connect to server\n" );
		ssh_close( info->socket );
		goto error;
	}

	if( sifs_mount( info ) < 0 ) {
		SIFS_DEBUG( "Unable to mount remote filesystem\n" );
		ssh_disconnect( info->socket );
		ssh_close( info->socket );
		goto error;
	}

	return 0;

error:
	info->socket = NULL;
	return -1;
}



/*
 *	Build path from dentries
 */

int sifs_getname( struct dentry* dentry, char* name )
{
	int len = 0;
	struct dentry* p = dentry;


SIFS_PROFILE( "Build path from dentries\n" );

	name[0] = '\0';
	if( dentry == dentry->d_parent ) {
		name[0] = '.';
		name[1] = '\0';
		return 0;
	}

	for( p = dentry; p != p->d_parent; p = p->d_parent ) {
		len += p->d_name.len + 1;
	}

	if( len > SIFS_MAXPATHLEN ) {
		SIFS_DEBUG( "Path to file is too long\n" );
		return -ENAMETOOLONG;
	}

	name[--len] = '\0';
	p = dentry;

	while( p->d_parent != p->d_parent->d_parent ) {
		len -= p->d_name.len;
		strncpy( &(name[len]), p->d_name.name, p->d_name.len );
		name[--len] = '/';
		p = p->d_parent;
	}

	len -= p->d_name.len;
	strncpy( &(name[len]), p->d_name.name, p->d_name.len );

	return 0;
}


/*
 * Whenever a lookup succeeds, we know the parent directories
 * are all valid, so we want to update the dentry timestamps.
 */
void sifs_renew_times( struct dentry* dentry )
{
SIFS_PROFILE( "Renew dentry times\n" );

	dentry->d_time = jiffies;

	while( !IS_ROOT( dentry ) ) {
		dentry = dentry->d_parent;
		dentry->d_time = jiffies;
	}
}
