/*
 *	Packet layer for communication to the kernel
 */



#include "packet.h"



/*
 *	Create a communication endpoint (a socket)
 */

struct psocket* packet_socket( int domain, int type, int protocol )
{
	struct psocket* psock = NULL;


	if( ( psock = (struct psocket*) malloc( sizeof( struct psocket ) ) ) == NULL ) {
		fprintf( stderr, "packet_socket: Out of memory\n" );
		errno = ENOMEM;
		return NULL;
	}

	memset( psock, 0, sizeof( struct psocket ) );

	if( ( psock->sockfd = socket( domain, type, protocol ) ) < 0 ) {
		perror( "packet_socket: Unable to create socket" );
		free( psock );
		return NULL;
	}

	return psock;
}



/*
 *	Terminate a connection
 */

int packet_close( struct psocket* psock )
{
	close( psock->sockfd );
	free( psock );
	return 0;
}



/*
 *	Bind the socket to an address and port
  */

int packet_bind( struct psocket* psock, struct sockaddr* myaddr, socklen_t addrlen )
{
	return bind( psock->sockfd, myaddr, addrlen );
}



/*
 *	Listen to the specified address and port
  */

int packet_listen( struct psocket* psock, int backlog )
{
	return listen( psock->sockfd, backlog );
}



/*
 *	Accept a connection
  */

struct psocket* packet_accept( struct psocket* lsock, struct sockaddr* addr, socklen_t* addrlen, char* uid, char* pwd )
{
	ssize_t size = 0;
	size_t packetlen = 0;
	size_t uidlen = 0;
	size_t pwdlen = 0;
	u_int8_t buffer[524];
	struct psocket* csock;


	memset( buffer, 0, 524 );

	if( ( csock = (struct psocket*) malloc( sizeof( struct psocket ) ) ) == NULL ) {
		fprintf( stderr, "packet_accept: Out of memory\n" );
		return NULL;
	}

	if( ( csock->sockfd = accept( lsock->sockfd, addr, addrlen ) ) < 0 ) {
		fprintf( stderr, "packet_accept: Accept new connection failed\n" );
		free( csock );
		return NULL;
	}

	// Read username and password from kernel to verify the client

	if( ( size = atomicio( read, csock->sockfd, buffer, 5 ) ) <= 0 ) {
		fprintf( stderr, "packet_accept: Read connect header failed\n" );
		free( csock );
		return NULL;
	}

	if( buffer[4] != 0 ) {   // Only a connect packet is allowed
		fprintf( stderr, "packet_accept: Invalid packet\n" );
		free( csock );
		errno = EPROTO;
		return NULL;
	}

	if( ( packetlen = GET_32BIT( buffer ) ) > 518 ) {
		fprintf( stderr, "packet_accept: Invalid packet length\n" );
		errno = EPROTO;
		return NULL;
	}

	if( ( size = atomicio( read, csock->sockfd, buffer + 5, packetlen ) ) <= 0 ) {
		fprintf( stderr, "packet_accept: Read connect payload failed\n" );
		free( csock );
		return NULL;
	}

	if( GET_16BIT( buffer + 5 ) != 2 ) {   // There must be a uid and pwd string
		fprintf( stderr, "packet_accept: Invalid payload\n" );
		free( csock );
		errno = EPROTO;
		return NULL;
	}

	uidlen = GET_16BIT( buffer + 7 );
	pwdlen = GET_16BIT( buffer + 7 + 2 + uidlen );

	if( uidlen != strlen( uid ) + 1 || pwdlen != strlen( pwd ) + 1 ) {
		fprintf( stderr, "packet_accept: Invalid credentials\n" );
		free( csock );
		errno = EACCES;
		return NULL;
	}

	if( memcmp( buffer + 9, uid, uidlen ) != 0 ) {
		fprintf( stderr, "packet_accept: Invalid username\n" );
		free( csock );
		errno = EACCES;
		return NULL;
	}

	if( memcmp( buffer + 9 + uidlen + 2, pwd, pwdlen ) != 0 ) {
		fprintf( stderr, "packet_accept: Invalid cookie\n" );
		free( csock );
		errno = EACCES;
		return NULL;
	}

	return csock;
}



/*
 *	Receive message from the kernel
 */

ssize_t packet_read( struct psocket* psock, void* buffer, size_t count )
{
	ssize_t size = 0;
	size_t packetlen = 0;
	u_int8_t header[5];


	//	There is more data available from the first packet

	if( psock->avail != 0 ) {
		size = psock->avail;

		if( count < (size_t) size ) {
			psock->avail -= count;
			size = count;
		}

		if( ( size = read( psock->sockfd, buffer, size ) ) <= 0 ) {
			perror( "packet_read: Reading remaining packet failed" );
			return size;
		}

		psock->avail = 0;
		return size;
	}


	//	No more data available from the previous packet, read a new one

	if( ( size = atomicio( read, psock->sockfd, header, 5 ) ) <= 0 ) {
		perror( "packet_read: Reading packet header failed" );
		return size;
	}

	if( header[4] == 1 ) {   // Disconnect and terminate
		return -2;
	}

	if( header[4] != 2 ) {   // Only data packets are allowed
		fprintf( stderr, "packet_read: Invalid packet type\n" );
		errno = EPROTO;
		return -1;
	}

	packetlen = GET_32BIT( header );
	size = packetlen;

	if( count < packetlen ) {
		psock->avail = packetlen - count;
		size = count;
	}

	if( ( size = read( psock->sockfd, buffer, size ) ) <= 0 ) {
		perror( "packet_read: Reading packet payload failed" );
		return size;
	}

	return size;
}



/*
 *	Send message to the kernel
 */

ssize_t packet_write( struct psocket* psock, void* buffer, size_t count )
{
	ssize_t size = 0;
	u_int8_t header[5];


	PUT_32BIT( header, count );
	header[4] = 2;   // data packet

	if( ( size = atomicio( write, psock->sockfd, header, 5 ) ) <= 0 ) {
		perror( "packet_write: Writing packet header failed" );
		return size;
	}

	if( ( size = atomicio( write, psock->sockfd, buffer, count ) ) <= 0 ) {
		perror( "packet_write: Writing packet payload failed" );
		return size;
	}

	return size;
}
