/*
 *	Packet layer for communication to the kernel
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include "atomicio.h"
#include "getput.h"


#ifndef PACKET_H
#define PACKET_H


#define SSH_BUFSIZE 8200



/*
 *	Struct for the packet socket interface
 */

struct psocket {
	int sockfd;
	u_int32_t avail;
};



/*
 *	Client sifsd packet protocol interface
 */


/*
 *	Create a communication endpoint (socket)
 *
 *	domain: used lowlevel protocol, e.g. PF_INET for IPv4
 *	type: type of transport protocol, e.g SOCK_STREAM for TCP
 *	protocol: specific protocol (default: 0)
 *
 * Returns a pointer to the packet struct on success or NULL on error
 */

struct psocket* packet_socket( int domain, int type, int protocol );


/*
 *	Terminate a connection
 *
 *	psock: allocated struct from packet_socket
 *
 *	Returns 0 on success and -1 on error
 */

int packet_close( struct psocket* psock );


/*
 *	Bind a socket to a local interface
 *
 *	psock: allocated struct from packet_socket
 *	myaddr: address and port where the socket should be bound to
 *	addrlen: size of myaddr
 *
 *	Returns 0 on success or -1 on error
 */

int packet_bind( struct psocket* psock, struct sockaddr* myaddr, socklen_t addrlen );


/*
 *	Listen to the specified address and port
 *
 *	psock: allocated struct from packet_socket
 *	backlog: number of allowed pending connections
 *
 *	Returns 0 on success or -1 on error
 */

int packet_listen( struct psocket* psock, int backlog );


/*
 *	Accept a connection
 *
 *	psock: allocated struct from packet_socket
 *	addr: pointer where the address and port of the other side is stored
 *	addrlen: size of addr (may be modified by the call)
 *
 * Returns a new struct psocket on success or -1 on error
 */

struct psocket* packet_accept( struct psocket* psock, struct sockaddr* addr, socklen_t* addrlen, char* uid, char* pwd );


/*
 *	Receive a message from the kernel
 *
 *	psock: allocated struct from packet_socket
 *	buffer: byte array where the message should be copied to
 *	count: size of the buffer array
 *
 *	Returns the number of read bytes on success and -1 on error
 */

ssize_t packet_read( struct psocket* psock, void* buffer, size_t count );


/*
 *	Send a message to the kernel
 *
 *	psock: allocated struct from packet_socket
 *	buffer: byte array which holds the message
 *	count: size of the buffer array
 *
 *	Returns the number of writtten bytes on success and -1 on error
 */

ssize_t packet_write( struct psocket* psock, void* buffer, size_t count );

#endif
