/*
 *	SIFS client daemon
 *
 * Establishes an authenticated and encrypted connection to an OpenSSH
 * server and forwards all data between the kernel filesystem driver and the
 * OpenSSH SIFS server module.
 *
 * (c) Norbert Sendetzky
 * norbert@linuxnetworks.de
 * Licenced under GPL
 */



#include "sifs-client.h"



void usage( char* progname )
{
	fprintf( stderr, "usage: %s -l loginname -p port host\n", progname );
	exit( 1 );
}



int read_cookie( struct parameter* param )
{
	int index = 0;


	fprintf( stdout, "Please enter a random cookie string: " );
	system( "stty -echo" );

	do {
		if( index >= 255 ) {
			fprintf( stderr, "Cookie may not exceed 255 character\n" );
			return -1;
		}
		param->cookie[index] = (char) getchar();
	}
	while( param->cookie[index++] != '\n' );

	param->cookie[--index] = '\0';   // \0 terminate the string
	system( "stty echo" );
	fprintf( stdout, "\n" );

	return 0;
}



int driver_listen( struct parameter* param )
{
	struct sockaddr_in myaddr;
	socklen_t socklen = sizeof( struct sockaddr_in );


	memset( &myaddr, 0, socklen );

	myaddr.sin_family = PF_INET;
	myaddr.sin_addr.s_addr = inet_addr( "127.0.0.1" );
	myaddr.sin_port = htons( param->port );

	if( ( param->lsock = packet_socket( PF_INET, SOCK_STREAM, 0 ) ) < 0 ) {
		fprintf( stderr, "Unable to create new socket\n" );
		return -1;
	}

	if( packet_bind( param->lsock, (struct sockaddr*) &myaddr, socklen ) < 0 ) {
		perror( "Unable to bind socket to address" );
		return -1;
	}

	if( packet_listen( param->lsock, 1 ) < 0 ) {
		perror( "Unable to listen to socket" );
		return -1;
	}

	return 0;
}



int main( int argc, char* argv[] )
{
	pid_t pid;
	int status = 0;
	struct sockaddr_in myaddr;
	socklen_t socklen = sizeof( struct sockaddr_in );
	struct parameter param;


	if( argc != 6 ) {
		usage( argv[0] );
		return 1;
	}

	memset( &param, 0, sizeof( struct parameter ) );

	if( strlen( argv[2] ) > 255 && strlen( argv[5] ) > 255 ) {
		fprintf( stderr, "loginname or host string is longer than 255 character\n" );
		return 1;
	}

	strncpy( param.host, argv[5], 255 );
	strncpy( param.login, argv[2], 255 );

	param.port = (u_int16_t) strtoul( argv[4], NULL, 10 );
	if( errno == ERANGE ) {
		fprintf( stderr, "Invalid port number\n" );
		return 1;
	}

	if( read_cookie( &param ) < 0 ) {
		fprintf( stderr, "Unable to read cookie\n" );
		return 1;
	}
SIFS_PROFILE( "main: Cookie for authentication read\n" );

	if( driver_listen( &param ) < 0 ) {
		fprintf( stderr, "Unable to listen for filesystem driver connection\n" );
		return 1;
	}
SIFS_PROFILE( "main: Creating listening socket for kernel succeeded\n" );

	while( 1 ) {
SIFS_PROFILE( "main: Waiting for kernel filesystem driver\n" );

		if( ( param.csock = packet_accept( param.lsock, (struct sockaddr*) &myaddr, &socklen, argv[2], param.cookie ) ) == NULL ) {
			fprintf( stderr, "Unable to get connection from kernel\n" );
			return -1;
		}
SIFS_PROFILE( "main: Accepted connection from kernel\n" );

		if( ( pid = fork() ) < 0 ) {
			perror( "Unable to fork child" );
			return -1;
		}

		if( pid == 0 ) {
			if( packet_close( param.lsock ) < 0 ) {
				fprintf( stderr, "Closeing listen socket failed\n" );
				exit( 1 );
			}
SIFS_PROFILE( "main: Closed listening socket in client process\n" );

			if( server_transport( &param ) < 0 ) {
				exit( 1 );
			}
			exit ( 2 );
		}

		if( packet_close( param.csock ) < 0 ) {
			fprintf( stderr, "Closing new kernel socket failed\n" );
			return -1;
		}
SIFS_PROFILE( "main: Closed kernel connection in parent process\n" );

		if( wait( &status ) == -1 ) {
			fprintf( stderr, "Unable to wait for process\n" );
			return -1;
		}

		if( WEXITSTATUS( status ) == 2 ) {
			return 0;
		}
SIFS_PROFILE( "main: Waiting for reconnect\n" );
	}
SIFS_PROFILE( "main: Terminate normally\n" );

	return 0;
}
