/*
 *	SIFS client daemon
 *
 * It establishes an authenticated and encrypted connection to an OpenSSH
 * server and forwards all data between the kernel filesystem driver and the
 * OpenSSH SIFS server module.
 *
 * (c) Norbert Sendetzky
 * norbert@linuxnetworks.de
 * Licenced under GPL
 */



#include "sifs-client.h"



int server_connect( struct parameter* param )
{
SIFS_PROFILE( "server_connect: Connect to server\n" );

	if( socketpair( AF_UNIX, SOCK_STREAM, 0, param->socketpair ) < 0 ) {
		perror( "Unable to create socketpair" );
		return -1;
	}
SIFS_PROFILE( "server_connect: Created socketpair for interprocess communication\n" );

	if( ( param->pid = fork() ) < 0 ) {
		perror( "Unable to fork child" );
		return -1;
	}

	if( param->pid == 0 ) {
		char* sshargv[] = { "/usr/bin/ssh", "-l", NULL, NULL, "-o", "ForwardX11=no", "-o", "ForwardAgent=no", "-o", "Protocol=2", "-s", "sifs", NULL };

		sshargv[2] = param->login;
		sshargv[3] = param->host;

		close( param->socketpair[0] );

		if( packet_close( param->csock ) < 0 ) {
			fprintf( stderr, "Closing socket to kernel failed\n" );
			exit( 1 );
		}
SIFS_PROFILE( "server_connect: Closed unused sockets in child process\n" );

		if( dup2( param->socketpair[1], 0 ) < 0 ) {	/* stdin */
			perror( "Unable to redirect stdin" );
			exit( 1 );
		}

		if( dup2( param->socketpair[1], 1 ) < 0 ) {	/* stdout - same socket as stdin. */
			perror( "Unable to redirect stdout" );
			exit( 1 );
		}
SIFS_PROFILE( "server_connect: Redirected stdin and stdout to sockets in child process\n" );

		execv( "/usr/bin/ssh", sshargv );
		fprintf( stderr, "Unable to execute server program\n" );
		exit( 1 );
	}

	close( param->socketpair[1] );
SIFS_PROFILE( "server_connect: Forking client process suceeded\n" );

	return 0;
}



int server_transport( struct parameter* param )
{
	fd_set rfds;
	int maxfd = 0;
	ssize_t size = 0;
	u_int8_t buffer[8200];


SIFS_PROFILE( "server_transport: Forward packets back and forth\n" );

	if( server_connect( param ) < 0 ) {
		fprintf( stderr, "Unable to connect to server\n" );
		return -1;
	}

	maxfd = param->socketpair[0];
	if( param->csock->sockfd > param->socketpair[0] ) {
		maxfd = param->csock->sockfd;
	}
	maxfd++;

	FD_ZERO( &rfds );
	FD_SET( param->csock->sockfd, &rfds );
	FD_SET( param->socketpair[0], &rfds );

	while( select( maxfd, &rfds, NULL, NULL, NULL ) != -1 )
	{
SIFS_PROFILE( "server_transport: Select signalized data for reading\n" );
		if( FD_ISSET( param->csock->sockfd, &rfds ) )
		{
SIFS_PROFILE( "server_transport: Data from kernel filesystem driver available\n" );
			if( ( size = packet_read( param->csock, buffer, 8200 ) ) <= 0 ) {
//				perror( "server_transport: Unable to read data from filesystem driver" );
				goto error;
			}
SIFS_PROFILE( "server_transport: Read data from kernel sucessfully\n" );

			fprintf( stderr, "Request:\n" );
			fprintf( stderr, "- Sequence number: %d\n", GET_32BIT( buffer ) );
			fprintf( stderr, "- Payload length: %d\n", GET_32BIT( buffer + 4 ) );
			fprintf( stderr, "- Opcode: %d\n", GET_32BIT( buffer + 8 ) );
			fprintf( stderr, "\n" );

			if( ( size = write( param->socketpair[0], buffer, size ) ) <= 0 ) {
				perror( "server_transport: Unable to write data to server module" );
				goto error;
			}
SIFS_PROFILE( "server_transport: Wrote data to server sucessfully\n" );
		}

		if( FD_ISSET( param->socketpair[0], &rfds ) )
		{
SIFS_PROFILE( "server_transport: Data from server available\n" );
			if( ( size = read( param->socketpair[0], buffer, 8200 ) ) <= 0 ) {
//				perror( "server_transport: Unable to read data from server" );
				goto error;
			}
SIFS_PROFILE( "server_transport: Read data from server sucessfully\n" );

			fprintf( stderr, "Reply:\n" );
			fprintf( stderr, "- Sequence number: %d\n", GET_32BIT( buffer ) );
			fprintf( stderr, "- Payload length: %d\n", GET_32BIT( buffer + 4 ) );
			fprintf( stderr, "- Error code: %d\n", GET_32BIT( buffer + 8 ) );
			fprintf( stderr, "\n" );

			if( ( size = packet_write( param->csock, buffer, size ) ) <= 0 ) {
				perror( "server_transport: Unable to write data to kernel filesystem driver" );
				goto error;
			}
SIFS_PROFILE( "server_transport: Wrote data to kernel sucessfully\n" );
		}

		FD_ZERO( &rfds );
		FD_SET( param->csock->sockfd, &rfds );
		FD_SET( param->socketpair[0], &rfds );
	}

	fprintf( stderr, "Select call failed\n" );

error:
SIFS_PROFILE( "server_transport: Connection closed by the other side\n" );
	close( param->socketpair[0] );

	if( packet_close( param->csock ) < 0 ) {
		fprintf( stderr, "Closing socket to kernel failed\n" );
	}
SIFS_PROFILE( "server_transport: Closed sockets sucessfully\n" );

	waitpid( param->pid, NULL, 0 );

	if( size == -2 ) {
SIFS_PROFILE( "server_transport: Kernel sent disconnect message\n" );
		return 0;
	}
SIFS_PROFILE( "server_transport: Kernel sent no disconnect message\n" );

return 0;
	return -1;
}

