/*
 * Returns the requested content of a file
 */



#include "sifs-process.h"



int sifs_read( struct connection* conn )
{
	int fd = 0;
	size_t len = 0;
	loff_t size = 0;
	loff_t offset = 0;
	ssize_t rbytes = 0;
	char* path = NULL;
	int status = SIFS_EOK;
	u_int32_t packetlen = GET_32BIT( conn->in.buffer + 4 );


SIFS_PROFILE( "sifs_read: Read partitial file content:\n" );

	memcpy( conn->out.buffer, conn->in.buffer, 4 );

	if( packetlen > 1044 ) {   // offset + size + one string
		if( consume_packet( conn->in.fd, conn->in.buffer, MAXBUFSIZE, packetlen ) < 0 ) {
			SIFS_DEBUG( "sifs_read: Unable to consume packet\n" );
			return -1;
		}
		SIFS_DEBUG( "sifs_read: Packet is longer than allowed\n" );
		status = SIFS_EPROTO;
		goto out;
	}
SIFS_PROFILE( "sifs_read: Request payload length is ok\n" );

	if( atomicio( read, conn->in.fd, conn->in.buffer, packetlen ) < 0 ) {
		SIFS_DEBUG( "sifs_read: Receiving request payload from client failed\n" );
		return -1;
	}
SIFS_PROFILE( "sifs_read: Receiving request payload succeeded\n" );

	if( conn->mntpath[0] == 0 ) {   // No filesystem is mounted
		SIFS_DEBUG( "sifs_read: There was no mount request before\n" );
		status = SIFS_EPROTO;
		goto out;
	}
SIFS_PROFILE( "sifs_read: Working directory is correct\n" );

	offset = GET_64BIT( conn->in.buffer );
	size = GET_64BIT( conn->in.buffer + 8 );

	if( GET_16BIT( conn->in.buffer + 16 ) != 1 ) {   // More than one string
		SIFS_DEBUG( "sifs_read: Request contains more strings than allowed\n" );
		status = SIFS_EPROTO;
		goto out;
	}
SIFS_PROFILE( "sifs_read: Number of request strings is ok\n" );

	if( ( len = str_decode( conn->in.buffer + 18, &path ) ) < 0 ) {
		SIFS_DEBUG( "sifs_read: Decoding string failed\n" );
		status = len;
		goto out;
	}
SIFS_PROFILE( "sifs_read: Decoding string succeeded, path = " );
SIFS_PROFILE( path );
SIFS_PROFILE( "\n" );

	if( ( fd = desc_get( path ) ) < 0 ) {
SIFS_PROFILE( "sifs_read: File was not opened before\n" );

		if( ( fd = open( path, O_RDWR ) ) < 0 ) {
			SIFS_DEBUG( "sifs_read: open syscall failed\n" );
			if( errno == ENAMETOOLONG ) {
				status = SIFS_ENAMETOOLONG;
				goto out;
			}

			if( errno == ENOENT || errno == ENOTDIR || errno == EISDIR ) {
				status = SIFS_ENOENT;
				goto out;
			}

			status = SIFS_ESERVER;
			goto out;
		}
SIFS_PROFILE( "sifs_read: open syscall succeeded\n" );

		if( desc_insert( path, fd ) < 0 ) {
			SIFS_DEBUG( "sifs_read: Inserting file descriptor into hash table failed\n" );
			status = SIFS_ESERVER;
			goto out;
		}
SIFS_PROFILE( "sifs_read: Inserting file descriptor into hash table succeeded\n" );

		if( desc_increase( path ) < 0 ) {
			SIFS_DEBUG( "sifs_read: Increasing reference counter failed\n" );
			status = SIFS_ESERVER;
			goto out;
		}
SIFS_PROFILE( "sifs_read: Inserting reference counter succeeded\n" );
	}

	if( lseek( fd, offset, SEEK_SET ) < 0 ) {
		SIFS_DEBUG( "sifs_read: lseek syscall failed\n" );
		status = SIFS_ESERVER;
		goto out;
	}
SIFS_PROFILE( "sifs_read: Inserting reference counter succeeded\n" );

	len = size;
	if( size > MAXBUFSIZE - 20 ) {
		len = MAXBUFSIZE - 20;
	}

	if( ( size = read( fd, conn->out.buffer + 20, len ) ) < 0 ) {
		SIFS_DEBUG( "sifs_read: read syscall failed\n" );
		if( errno == EIO ) {
			status = SIFS_EIO;
			goto out;
		}

		if( errno == EINVAL ) {
			status = SIFS_EINVAL;
			goto out;
		}

		status = SIFS_ESERVER;
		goto out;
	}
SIFS_PROFILE( "sifs_read: read syscall succeeded\n" );
fprintf( stderr, "read bytes = %d\n", size );

	PUT_64BIT( conn->out.buffer + 12, size );
	size += 8;
	status = SIFS_EOK;

out:
	PUT_32BIT( conn->out.buffer + 4, size );
	PUT_32BIT( conn->out.buffer + 8, status );

	if( atomicio( write, conn->out.fd, conn->out.buffer, size + 12 ) < 0 ) {
		SIFS_DEBUG( "sifs_read: Sending reply to client failed\n" );
		return -1;
	}
SIFS_PROFILE( "sifs_read: Sending reply to client succeeded\n" );

	return 0;
}

