/*
 * Returns all file names stored in a directory
 */



#include "sifs-process.h"



int sifs_readdir( struct connection* conn )
{
	int i, j, len;
	int entries = 0;
	char* path = NULL;
	int status = SIFS_EOK;
	struct dirent** namelist;
	u_int32_t size = 0;
	u_int32_t totlen = 0;
	u_int32_t packetsize = 0;
	u_int32_t packetlen = GET_32BIT( conn->in.buffer + 4 );


SIFS_PROFILE( "sifs_readdir: Read directory content:\n" );

	memcpy( conn->out.buffer, conn->in.buffer, 4 );

	if( packetlen > 1028 ) {   // one string
		if( consume_packet( conn->in.fd, conn->in.buffer, MAXBUFSIZE, packetlen ) < 0 ) {
			SIFS_DEBUG( "sifs_readdir: Unable to consume packet\n" );
			return -1;
		}
		SIFS_DEBUG( "sifs_readdir: Packet is longer than allowed\n" );
		status = SIFS_EPROTO;
		goto error;
	}
SIFS_PROFILE( "sifs_readdir: Request payload length is ok\n" );

	if( atomicio( read, conn->in.fd, conn->in.buffer, packetlen ) < 0 ) {
		SIFS_DEBUG( "sifs_readdir: Reading request payload from client failed\n" );
		return -1;
	}
SIFS_PROFILE( "sifs_readdir: Receiving request payload succeeded\n" );

	if( conn->mntpath[0] == 0 ) {   // No filesystem is mounted
		SIFS_DEBUG( "sifs_readdir: There was no mount request before\n" );
		status = SIFS_EPROTO;
		goto error;
	}
SIFS_PROFILE( "sifs_readdir: Working directory is correct\n" );

	if( GET_16BIT( conn->in.buffer ) != 1 ) {   // More than one string
		SIFS_DEBUG( "sifs_readdir: Request contains more strings than allowed\n" );
		status = SIFS_EPROTO;
		goto error;
	}
SIFS_PROFILE( "sifs_readdir: Number of request strings is ok\n" );

	if( ( len = str_decode( conn->in.buffer + 2, &path ) ) < 0 ) {
		SIFS_DEBUG( "sifs_readdir: Decoding string failed\n" );
		status = len;
		goto error;
	}
SIFS_PROFILE( "sifs_readdir: Decoding string succeeded, path = " );
SIFS_PROFILE( path );
SIFS_PROFILE( "\n" );

	if( ( entries= scandir( path, &namelist, 0, 0 ) ) < 0 ) {
		SIFS_DEBUG( "sifs_readdir: scandir syscall failed\n" );
		if( errno == ENAMETOOLONG ) {
			status = SIFS_ENAMETOOLONG;
			goto error;
		}

		if( errno == ENOENT || errno == ENOTDIR ) {
			status = SIFS_ENOENT;
			goto error;
		}

		if( errno == EACCES ) {
			status = SIFS_EACCES;
			goto error;
		}

		status = SIFS_ESERVER;
		goto error;
	}
SIFS_PROFILE( "sifs_readdir: scandir syscall succeeded\n" );

	for( i = 0; i < entries; i++ ) {
		size = strlen( namelist[i]->d_name ) + 1;

		if( size + totlen < 0xFFFFFFFF ) {   // prevent wrap arounds
			totlen += size;
		} else {   // discard all other strings
			for( j = i; j < entries; j++ ) {
				SIFS_DEBUG( "sifs_readdir: Discard string, string = " );
				SIFS_DEBUG( namelist[j] );
				SIFS_DEBUG( "\n" );
				free( namelist[j] );
			}
			entries = i - 1;
		}
	}
SIFS_PROFILE( "sifs_readdir: Directory content ready for sending\n" );

	PUT_32BIT( conn->out.buffer + 4, 2 + 2 * entries + totlen );
	PUT_32BIT( conn->out.buffer + 8, SIFS_EOK );
	PUT_16BIT( conn->out.buffer + 12, entries );

	i = 0;
	packetsize = 12 + 2;
	while( i < entries ) {
		if( packetsize + 2 + strlen( namelist[i]->d_name ) + 1 > MAXBUFSIZE ) {
			if( atomicio( write, conn->out.fd, conn->out.buffer, packetsize ) < 0 ) {
				SIFS_DEBUG( "sifs_readdir: Sending partitial reply to client failed\n" );
				return -1;
			}
SIFS_PROFILE( "sifs_readdir: Sending partitial reply to client succeeded\n" );
			packetsize = 0;
		}

		if( ( len = str_encode( conn->out.buffer + packetsize, namelist[i]->d_name ) ) < 0 ) {
			SIFS_DEBUG( "sifs_readdir: Encoding string failed\n" );
			status = len;
			goto error;
		}
SIFS_PROFILE( "sifs_readdir: Encoding string succeeded\n" );
		free( namelist[i] );
		packetsize += len;
		i++;
	}

	free( namelist );

	if( atomicio( write, conn->out.fd, conn->out.buffer, packetsize ) < 0 ) {
		SIFS_DEBUG( "sifs_readdir: Sending last partitial reply to client failed\n" );
		return -1;
	}
SIFS_PROFILE( "sifs_readdir: Sending last partitial reply to client succeeded\n" );

	return 0;

error:
	PUT_32BIT( conn->out.buffer + 4, 0 );
	PUT_32BIT( conn->out.buffer + 8, status );

	if( atomicio( write, conn->out.fd, conn->out.buffer, 12 ) < 0 ) {
		SIFS_DEBUG( "sifs_readdir: Sending error reply to client failed\n" );
		return -1;
	}
SIFS_PROFILE( "sifs_readdir: Sending error reply to client succeeded\n" );

	return 0;
}
