/*
 * Return various information about the underlying filesystem
 */



#include "sifs-process.h"



int sifs_statfs( struct connection* conn )
{
	int len = 0;
	int size = 0;
	struct statfs stat;
	char* path = NULL;
	int status = SIFS_EOK;
	u_int32_t  packetlen = GET_32BIT( conn->in.buffer + 4 );


SIFS_PROFILE( "sifs_statfs: Stating filesystem:\n" );

	memcpy( conn->out.buffer, conn->in.buffer, 4 );

	if( packetlen > 1028 ) {   // one string
		if( consume_packet( conn->in.fd, conn->in.buffer, MAXBUFSIZE, packetlen ) < 0 ) {
			SIFS_DEBUG( "sifs_statfs: Unable to consume packet\n" );
			return -1;
		}
		SIFS_DEBUG( "sifs_statfs: Packet is longer than allowed\n" );
		status = SIFS_EPROTO;
		goto out;
	}
SIFS_PROFILE( "sifs_statfs: Request payload length is ok\n" );

	if( atomicio( read, conn->in.fd, conn->in.buffer, packetlen ) < 0 ) {
		SIFS_DEBUG( "sifs_statfs: Reading request payload from client failed\n" );
		return -1;
	}
SIFS_PROFILE( "sifs_statfs: Receiving request payload succeeded\n" );

	if( conn->mntpath[0] == 0 ) {   // No filesystem is mounted
		SIFS_DEBUG( "sifs_statfs: There was no mount request before\n" );
		status = SIFS_EPROTO;
		goto out;
	}
SIFS_PROFILE( "sifs_statfs: Working directory is correct\n" );

	if( GET_16BIT( conn->in.buffer ) != 1 ) {   // More than one string
		SIFS_DEBUG( "sifs_statfs: Request contains more strings than allowed\n" );
		status = SIFS_EPROTO;
		goto out;
	}
SIFS_PROFILE( "sifs_statfs: Number of request strings is ok\n" );

	if( ( len = str_decode( conn->in.buffer + 2, &path ) ) < 0 ) {
		SIFS_DEBUG( "sifs_statfs: Decoding string failed\n" );
		status = len;
		goto out;
	}
SIFS_PROFILE( "sifs_statfs: Decoding string succeeded, path = " );
SIFS_PROFILE( path );
SIFS_PROFILE( "\n" );

	if( statfs( path, &stat ) < 0 ) {
		SIFS_DEBUG( "sifs_statfs: statfs syscall failed\n" );
		if( errno == ENAMETOOLONG ) {
			status = SIFS_ENAMETOOLONG;
			goto out;
		}

		if( errno == ENOENT || errno == ENOTDIR ) {
			status = SIFS_ENOENT;
			goto out;
		}

		if( errno == EACCES ) {
			status = SIFS_EACCES;
			goto out;
		}

		if( errno == EIO ) {
			status = SIFS_EIO;
			goto out;
		}

		if( errno == ENOSYS ) {
			status = SIFS_ENOTSUP;
			goto out;
		}

		status = SIFS_ESERVER;
		goto out;
	}
SIFS_PROFILE( "sifs_statfs: statfs syscall succeeded\n" );

	PUT_32BIT( conn->out.buffer + 12, stat.f_bsize );
	PUT_32BIT( conn->out.buffer + 16, stat.f_blocks );
	PUT_32BIT( conn->out.buffer + 20, stat.f_bfree );
	PUT_32BIT( conn->out.buffer + 24, stat.f_bavail );
	PUT_32BIT( conn->out.buffer + 28, stat.f_files );
	PUT_32BIT( conn->out.buffer + 32, stat.f_ffree );
	PUT_32BIT( conn->out.buffer + 36, stat.f_namelen );

	size = 28;

out:
	PUT_32BIT( conn->out.buffer + 4, size );
	PUT_32BIT( conn->out.buffer + 8, status );

	if( atomicio( write, conn->out.fd, conn->out.buffer, size + 12 ) < 0 ) {
		SIFS_DEBUG( "sifs_statfs: Sending reply to client failed\n" );
		return -1;
	}
SIFS_PROFILE( "sifs_statfs: Sending reply to client succeeded\n" );

	return 0;
}
